# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The leader script (of the leader/worker pair) for running jobs."""

import base64
import glob
import logging
import os
import pickle
import sys
import time
from typing import Any, Dict, List, Optional, Set, Union

import enlighten

from toil import resolveEntryPoint
from toil.batchSystems import DeadlockException
from toil.batchSystems.abstractBatchSystem import (AbstractBatchSystem,
                                                   BatchJobExitReason)
from toil.bus import (JobAnnotationMessage,
                      JobCompletedMessage,
                      JobFailedMessage,
                      JobIssuedMessage,
                      JobMissingMessage,
                      JobUpdatedMessage,
                      QueueSizeMessage)
from toil.common import Config, Toil, ToilMetrics
from toil.cwl.utils import CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE
from toil.job import (CheckpointJobDescription,
                      JobDescription,
                      ServiceJobDescription,
                      TemporaryID)
from toil.jobStores.abstractJobStore import (AbstractJobStore,
                                             NoSuchFileException,
                                             NoSuchJobException)
from toil.lib.throttle import LocalThrottle
from toil.provisioners.abstractProvisioner import AbstractProvisioner
from toil.provisioners.clusterScaler import ScalerThread
from toil.serviceManager import ServiceManager
from toil.statsAndLogging import StatsAndLogging
from toil.toilState import ToilState
from toil.exceptions import FailedJobsException

logger = logging.getLogger(__name__)

###############################################################################
# Implementation Notes
#
# Multiple predecessors:
#   There is special-case handling for jobs with multiple predecessors as a
#   performance optimization. This minimize number of expensive loads of
#   JobDescriptions from jobStores.  However, this special case could be unnecessary.
#   The JobDescription is loaded to update predecessorsFinished, in
#   _checkSuccessorReadyToRunMultiplePredecessors, however it doesn't appear to
#   write the JobDescription back the jobStore.  Thus predecessorsFinished may really
#   be leader state and could moved out of the JobDescription.  This would make this
#   special-cases handling unnecessary and simplify the leader.
#   Issue #2136
###############################################################################


class Leader:
    """
    Represents the Toil leader.

    Responsible for determining what jobs are ready to be scheduled, by
    consulting the job store, and issuing them in the batch system.
    """

    def __init__(self,
                 config: Config,
                 batchSystem: AbstractBatchSystem,
                 provisioner: Optional[AbstractProvisioner],
                 jobStore: AbstractJobStore,
                 rootJob: JobDescription,
                 jobCache: Optional[Dict[Union[str, TemporaryID], JobDescription]] = None) -> None:
        """
        Create a Toil Leader object.

        If jobCache is passed, it must be a dict from job ID to pre-existing
        JobDescription objects. Jobs will be loaded from the cache (which can be
        downloaded from the jobStore in a batch) when loading into the ToilState object.

        :param config:      A Config object holding the original user options/settings/args.
        :param batchSystem: The Batch System object containing functions to communicate with the
                            specific batch system (run jobs, check jobs, etc.).
        :param provisioner: Only for environments where we need to provision our own compute
                            resources prior to launching jobs, for example, spin up an AWS instance.
                            This is a Provisioner object with functions to create/remove resources.
        :param jobStore:    The Jobstore object for storage, containing functions to talk to a
                            central location where we store all of our files (jobs, inputs, etc.).
        :param rootJob:     The first job of the workflow that the leader will run.
        """
        # Object containing parameters for the run
        self.config = config

        # The job store
        self.jobStore = jobStore
        self.jobStoreLocator = config.jobStore

        # The ToilState will be the authority on the current state of the jobs
        # in the jobStore, and its bus is the one true place to listen for
        # state change information about jobs.
        self.toilState = ToilState(self.jobStore)

        if self.config.write_messages is not None:
            # Message bus messages need to go to the given file.
            # Keep a reference to the return value so the listener stays alive.
            self._message_subscription = self.toilState.bus.connect_output_file(self.config.write_messages)

        # Connect to the message bus, so we will get all the messages of these
        # types in an inbox.
        self._messages = self.toilState.bus.connect([JobUpdatedMessage])

        # Connect the batch system to the bus so it can e.g. annotate jobs
        batchSystem.set_message_bus(self.toilState.bus)

        # Load the jobs into the ToilState, now that we are able to receive any
        # resulting messages.
        # TODO: Give other components a chance to connect to the bus before
        # this, somehow, so they can also see messages from this?
        self.toilState.load_workflow(rootJob, jobCache=jobCache)

        logger.debug("Found %s jobs to start and %i jobs with successors to run",
                     self._messages.count(JobUpdatedMessage), len(self.toilState.successorCounts))

        # Batch system
        self.batchSystem = batchSystem
        assert len(self.batchSystem.getIssuedBatchJobIDs()) == 0  # Batch system must start with no active jobs!
        logger.debug("Checked batch system has no running jobs and no updated jobs")

        # Map of batch system IDs to job store IDs
        self.issued_jobs_by_batch_system_id: Dict[int, str] = {}

        # Number of preemptible jobs currently being run by batch system
        self.preemptibleJobsIssued = 0

        # Tracking the number service jobs issued,
        # this is used limit the number of services issued to the batch system
        self.serviceJobsIssued = 0
        self.serviceJobsToBeIssued: List[str] = [] # A queue of IDs of service jobs that await scheduling
        # Equivalents for service jobs to be run on preemptible nodes
        self.preemptibleServiceJobsIssued = 0
        self.preemptibleServiceJobsToBeIssued: List[str] = []

        # Timing of the rescuing method
        self.timeSinceJobsLastRescued = None

        # For each issued job's batch system ID, how many times did we not see
        # it when we should have? If this hits a threshold, the job is declared
        # missing and killed and possibly retried.
        self.reissueMissingJobs_missingHash: Dict[int, int] = {}

        # Class used to create/destroy nodes in the cluster, may be None if
        # using a statically defined cluster
        self.provisioner = provisioner

        # Create cluster scaling thread if the provisioner is not None
        self.clusterScaler = None
        if self.provisioner is not None and self.provisioner.hasAutoscaledNodeTypes():
            self.clusterScaler = ScalerThread(self.provisioner, self, self.config)

        # A service manager thread to start and terminate services
        self.serviceManager = ServiceManager(jobStore, self.toilState)

        # A thread to manage the aggregation of statistics and logging from the run
        self.statsAndLogging = StatsAndLogging(self.jobStore, self.config)

        # Set used to monitor deadlocked jobs
        self.potentialDeadlockedJobs: Set[str] = set()
        self.potentialDeadlockTime = 0

        # A dashboard that runs on the leader node in AWS clusters to track the state
        # of the cluster
        self.toilMetrics: Optional[ToilMetrics] = None

        # internal jobs we should not expose at top level debugging
        self.debugJobNames = ("CWLJob", "CWLWorkflow", "CWLScatter", "CWLGather",
                              "ResolveIndirect")

        self.deadlockThrottler = LocalThrottle(self.config.deadlockCheckInterval)

        self.statusThrottler = LocalThrottle(self.config.statusWait)

        # Control how often to poll the job store to see if we have been killed
        self.kill_throttler = LocalThrottle(self.config.kill_polling_interval)

        # For fancy console UI, we use an Enlighten counter that displays running / queued jobs
        # This gets filled in in run() and updated periodically.
        self.progress_overall = None
        # Also count failed/killed jobs
        self.progress_failed = None
        # Assign colors for them
        self.GOOD_COLOR = (0, 60, 108)
        self.BAD_COLOR = (253, 199, 0)
        # And set a format that shows failures
        self.PROGRESS_BAR_FORMAT = ('{desc}{desc_pad}{percentage:3.0f}%|{bar}| {count:{len_total}d}/{total:d} '
                                    '({count_1:d} failures) [{elapsed}<{eta}, {rate:.2f}{unit_pad}{unit}/s]')
        # TODO: No way to set background color on the terminal for the bar.

        # What exit code should the process use if the workflow failed?
        # Needed in case a worker detects a CWL issue that a CWL runner must
        # report to its caller.
        self.recommended_fail_exit_code = 1

    def run(self) -> Any:
        """
        Run the leader process to issue and manage jobs.

        :raises: toil.exceptions.FailedJobsException if failed jobs remain after running.

        :return: The return value of the root job's run function.
        """
        self.jobStore.write_kill_flag(kill=False)

        with enlighten.get_manager(stream=sys.stderr, enabled=not self.config.disableProgress) as manager:
            # Set up the fancy console UI if desirable
            self.progress_overall = manager.counter(total=0, desc='Workflow Progress', unit='jobs',
                                                    color=self.GOOD_COLOR, bar_format=self.PROGRESS_BAR_FORMAT)
            self.progress_failed = self.progress_overall.add_subcounter(self.BAD_COLOR)

            # Start the stats/logging aggregation thread
            self.statsAndLogging.start()
            if self.config.metrics:
                self.toilMetrics = ToilMetrics(self.toilState.bus, provisioner=self.provisioner)

            try:

                # Start service manager thread
                self.serviceManager.start()
                try:

                    # Create cluster scaling processes if not None
                    if self.clusterScaler is not None:
                        self.clusterScaler.start()

                    try:
                        # Run the main loop
                        self.innerLoop()
                    finally:
                        if self.clusterScaler is not None:
                            logger.debug('Waiting for workers to shutdown.')
                            startTime = time.time()
                            self.clusterScaler.shutdown()
                            logger.debug('Worker shutdown complete in %s seconds.', time.time() - startTime)

                finally:
                    # Ensure service manager thread is properly shutdown
                    self.serviceManager.shutdown()

            finally:
                # Ensure the stats and logging thread is properly shutdown
                self.statsAndLogging.shutdown()
                if self.toilMetrics:
                    self.toilMetrics.shutdown()

            # Filter the failed jobs
            self.toilState.totalFailedJobs = [j for j in self.toilState.totalFailedJobs if self.toilState.job_exists(j)]

            try:
                self.create_status_sentinel_file(self.toilState.totalFailedJobs)
            except OSError as e:
                logger.debug(f'Error from importFile with hardlink=True: {e}')

            logger.info("Finished toil run %s" %
                         ("successfully." if not self.toilState.totalFailedJobs \
                    else ("with %s failed jobs." % len(self.toilState.totalFailedJobs))))

            if len(self.toilState.totalFailedJobs):
                failed_jobs = []
                for job_id in self.toilState.totalFailedJobs:
                    # Refresh all the failed jobs to get e.g. the log file IDs that the workers wrote
                    self.toilState.reset_job(job_id)
                    failed_jobs.append(self.toilState.get_job(job_id))

                logger.info("Failed jobs at end of the run: %s", ' '.join(str(j) for j in failed_jobs))
                raise FailedJobsException(self.jobStore, failed_jobs, exit_code=self.recommended_fail_exit_code)

            return self.jobStore.get_root_job_return_value()

    def create_status_sentinel_file(self, fail: bool) -> None:
        """Create a file in the jobstore indicating failure or success."""
        logName = 'failed.log' if fail else 'succeeded.log'
        localLog = os.path.join(os.getcwd(), logName)
        open(localLog, 'w').close()
        self.jobStore.import_file('file://' + localLog, logName, hardlink=True)

        if os.path.exists(localLog):  # Bandaid for Jenkins tests failing stochastically and unexplainably.
            os.remove(localLog)

    def _handledFailedSuccessor(self, successor_id: str, predecessor_id: str) -> bool:
        """
        Deal with the successor having failed.

        :param successor_id: The successor which has failed.
        :param predecessor_id: The job which the successor comes after.
        :returns: True if there are still active successors.
                  False if all successors have failed and the job is queued to run to handle the failed successors.
        """
        logger.debug("Successor job: %s of job: %s has failed """
                     "predecessors", self.toilState.get_job(successor_id), self.toilState.get_job(predecessor_id))

        # Add the job to the set having failed successors
        self.toilState.hasFailedSuccessors.add(predecessor_id)

        # Reduce active successor count and remove the successor as an active successor of the job
        self.toilState.successor_returned(predecessor_id)
        self.toilState.successor_to_predecessors[successor_id].remove(predecessor_id)
        if len(self.toilState.successor_to_predecessors[successor_id]) == 0:
            self.toilState.successor_to_predecessors.pop(successor_id)

        # If the job now has no active successors, add to active jobs
        # so it can be processed as a job with failed successors.
        if self.toilState.count_pending_successors(predecessor_id) == 0:
            logger.debug("Job: %s has no successors to run "
                         "and some are failed, adding to list of jobs "
                         "with failed successors", self.toilState.get_job(predecessor_id))
            self._messages.publish(JobUpdatedMessage(predecessor_id, 0))
            # Report no successors are running
            return False
        else:
            # Some successors are still active
            return True

    def _checkSuccessorReadyToRunMultiplePredecessors(self, successor_id: str, predecessor_id: str) -> bool:
        """
        Check if a successor job is ready to run when there are multiple predecessors.

        :param successor_id: The successor which has failed.
        :param predecessor_id: The job which the successor comes after.

        :returns: True if the successor is ready to run now, and False otherwise.
        """
        # See implementation note at the top of this file for discussion of multiple predecessors

        # Remember that we have multiple predecessors, if we haven't already
        # TODO: do we ever consult this after building the ToilState?
        self.toilState.jobsToBeScheduledWithMultiplePredecessors.add(successor_id)

        # Grab the cached version of the job, where we have the in-memory finished predecessor info.
        successor = self.toilState.get_job(successor_id)

        # Grab the predecessor for reporting
        predecessor = self.toilState.get_job(predecessor_id)

        logger.debug("Successor job: %s of job: %s has multiple "
                     "predecessors", successor, predecessor)

        # Add the predecessor as a finished predecessor to the successor
        successor.predecessorsFinished.add(predecessor_id)
        # TODO: commit back???

        # If the successor is in the set of successors of failed jobs
        if successor_id in self.toilState.failedSuccessors:
            if not self._handledFailedSuccessor(successor_id, predecessor_id):
                # The job is not ready to run
                return False

        # If the successor job's predecessors have all not all completed then
        # ignore the successor as is not yet ready to run
        assert len(successor.predecessorsFinished) <= successor.predecessorNumber
        if len(successor.predecessorsFinished) == successor.predecessorNumber:
            # All the successor's predecessors are done now.
            # Remove the successor job from the set of waiting multi-predecessor jobs.
            self.toilState.jobsToBeScheduledWithMultiplePredecessors.remove(successor_id)
            return True
        else:
            # The job is not ready to run
            return False

    def _makeJobSuccessorReadyToRun(self, successor_id: str, predecessor_id: str) -> bool:
        """
        Make a successor job ready to run if possible.

        :param successor_id: The successor which should become ready.
        :param predecessor_id: The job which the successor comes after.
        :returns: False if the successor job should not yet be run or True otherwise.
        """
        #Build map from successor to predecessors.
        if successor_id not in self.toilState.successor_to_predecessors:
            self.toilState.successor_to_predecessors[successor_id] = set()
        assert isinstance(successor_id, str)
        assert isinstance(predecessor_id, str)
        self.toilState.successor_to_predecessors[successor_id].add(predecessor_id)

        # Grab the successor
        successor = self.toilState.get_job(successor_id)
        logger.debug("Added job %s as coming after job %s", successor, self.toilState.get_job(predecessor_id))
        if successor.predecessorNumber > 1:
            return self._checkSuccessorReadyToRunMultiplePredecessors(successor_id, predecessor_id)
        else:
            return True

    def _runJobSuccessors(self, predecessor_id: str) -> None:
        """
        Issue the successors of a job.

        :param predecessor_id: The ID of the job which the successors come after.
        """
        # TODO: rewrite!

        # Grab the predecessor's JobDescription
        predecessor = self.toilState.get_job(predecessor_id)

        # Grap the successors
        next_successors = predecessor.nextSuccessors()

        if next_successors is None or len(next_successors) == 0:
            raise RuntimeError(f"Job {self} trying to run successors, but it doesn't have any")
        logger.debug("Job: %s has %i successors to schedule",
                     predecessor_id, len(next_successors))
        #Record the number of successors that must be completed before
        #the job can be considered again
        assert self.toilState.count_pending_successors(predecessor_id) == 0, 'Attempted to schedule successors of the same job twice!'
        self.toilState.successors_pending(predecessor_id, len(next_successors))

        # For each successor schedule if all predecessors have been completed
        successors = []
        for successor_id in next_successors:
            try:
                successor = self.toilState.get_job(successor_id)
            except NoSuchJobException:
                # Job already done and gone, but probably shouldn't be. Or maybe isn't visible yet.
                # TODO: Shouldn't this be an error?
                logger.warning("Job %s is a successor of %s but is already done and gone.", successor_id, predecessor_id)
                # Don't try and run it
                continue
            if self._makeJobSuccessorReadyToRun(successor_id, predecessor_id):
                successors.append(successor)
        self.issueJobs(successors)

    def _processFailedSuccessors(self, predecessor_id: str):
        """
        Deal with some of a job's successors having failed.

        Either fail the job, or restart it if it has retries left and is a checkpoint
        job.
        """

        # Get the description
        predecessor = self.toilState.get_job(predecessor_id)

        if predecessor_id in self.toilState.servicesIssued:
            # The job has services running; signal for them to be killed.
            # Once they are killed, then the job will be updated again and then
            # scheduled to be removed.
            logger.warning("Telling job %s to terminate its services due to successor failure",
                           predecessor)
            self.serviceManager.kill_services(self.toilState.servicesIssued[predecessor_id],
                                              error=True)
        elif self.toilState.count_pending_successors(predecessor_id) > 0:
            # The job has non-service jobs running; wait for them to finish.
            # the job will be re-added to the updated jobs when these jobs
            # are done
            logger.debug("Job %s with ID: %s with failed successors still has successor jobs running",
                         predecessor, predecessor_id)
        elif (isinstance(predecessor, CheckpointJobDescription) and
              predecessor.checkpoint is not None and
              predecessor.remainingTryCount > 1):
            # If the job is a checkpoint and has remaining retries...
            # The logic behind using > 1 rather than > 0 here: Since this job has
            # been tried once (without decreasing its try count as the job
            # itself was successful), and its subtree failed, it shouldn't be retried
            # unless it has more than 1 try.
            if predecessor_id in self.toilState.jobs_issued:
                logger.debug('Checkpoint job %s was updated while issued', predecessor_id)
            else:
                # It hasn't already been reissued.
                # This check lets us be robust against repeated job update
                # messages (such as from services starting *and* failing), by
                # making sure that we don't stay in a state that where we
                # reissue the job every time we get one.
                logger.warning('Job: %s is being restarted as a checkpoint after the total '
                               'failure of jobs in its subtree.', predecessor_id)
                self.issueJob(predecessor)
        else:
            # Mark it totally failed
            logger.debug("Job %s is being processed as completely failed", predecessor_id)
            self.processTotallyFailedJob(predecessor_id)

    def _processReadyJob(self, job_id: str, result_status: int):
        # We operate on the JobDescription mostly.
        readyJob = self.toilState.get_job(job_id)

        logger.debug('Updating status of job %s with result status: %s',
                     readyJob, result_status)

        # TODO: Filter out nonexistent successors/services now, so we can tell
        # if they are all done and the job needs deleting?

        if self.serviceManager.services_are_starting(readyJob.jobStoreID):
            # This stops a job with services being issued by the serviceManager
            # from being considered further in this loop. This catch is
            # necessary because the job's service's can fail while being
            # issued, causing the job to be sent an update message. We don't
            # want to act on it; we want to wait until it gets the update it
            # gets when the service manager is done trying to start its
            # services.
            logger.debug("Got a job to update which is still owned by the service "
                         "manager: %s", readyJob.jobStoreID)
        elif readyJob.jobStoreID in self.toilState.hasFailedSuccessors:
            self._processFailedSuccessors(job_id)
        elif readyJob.command is not None or result_status != 0:
            # The job has a command it must be run before any successors.
            # Similarly, if the job previously failed we rerun it, even if it doesn't have a
            # command to run, to eliminate any parts of the stack now completed.
            isServiceJob = readyJob.jobStoreID in self.toilState.service_to_client

            # We want to run the job, and expend one of its "tries" (possibly
            # the initial one)

            # If the job has run out of tries or is a service job whose error flag has
            # been indicated, fail the job.
            if (readyJob.remainingTryCount == 0 or
                (isServiceJob and not self.jobStore.file_exists(readyJob.errorJobStoreID))):
                self.processTotallyFailedJob(job_id)
                logger.warning("Job %s is completely failed", readyJob)
            else:
                # Otherwise try the job again
                self.issueJob(readyJob)
        elif next(readyJob.serviceHostIDsInBatches(), None) is not None:
            # the job has services to run, which have not been started, start them
            # Build a map from the service jobs to the job and a map
            # of the services created for the job
            assert readyJob.jobStoreID not in self.toilState.servicesIssued
            self.toilState.servicesIssued[readyJob.jobStoreID] = set()
            for serviceJobList in readyJob.serviceHostIDsInBatches():
                for serviceID in serviceJobList:
                    assert serviceID not in self.toilState.service_to_client
                    self.toilState.reset_job(serviceID)
                    serviceHost = self.toilState.get_job(serviceID)
                    self.toilState.service_to_client[serviceID] = readyJob.jobStoreID
                    self.toilState.servicesIssued[readyJob.jobStoreID].add(serviceID)

            logger.debug("Giving job: %s to service manager to schedule its jobs", readyJob)
            # Use the service manager to start the services
            self.serviceManager.put_client(job_id)
        elif readyJob.nextSuccessors() is not None:
            # There are successors to run
            self._runJobSuccessors(job_id)
        elif readyJob.jobStoreID in self.toilState.servicesIssued:
            logger.debug("Telling job: %s to terminate its services due to the "
                         "successful completion of its successor jobs",
                         readyJob)
            self.serviceManager.kill_services(self.toilState.servicesIssued[readyJob.jobStoreID], error=False)
        else:
            # There are no remaining tasks to schedule within the job.
            #
            # We want it to go away. Usually the worker is responsible for
            # deleting completed jobs. But when the job has a
            # child/followOn/service dependents that have to finish first, it
            # gets kicked back here when they are done.
            #
            # We don't want to just issueJob() again because then we will
            # forget that this job is actually done and not a retry. A
            # successful-subtree checkpoint job can't be distinguished from a
            # being-retried one.
            #
            # So we do the cleanup here, to avoid reserving any real resources.
            # See: https://github.com/DataBiosphere/toil/issues/4158 and
            # https://github.com/DataBiosphere/toil/issues/3188
            #
            # TODO: Does this want to happen in another thread? What if we
            # don't manage to do all the jobstore file deletes and need to
            # retry them? What if there are 1000 of them?

            if readyJob.remainingTryCount > 0:
                # add attribute to let issueJob know that this is an empty job and should be deleted
                logger.debug("Job: %s is empty, we are cleaning it up", readyJob)

                try:
                    self.toilState.delete_job(readyJob.jobStoreID)
                except Exception as e:
                    logger.exception("Re-processing success for job we could not remove: %s", readyJob)
                    # Kick it back to being handled as succeeded again. We
                    # don't want to have a failure here cause a Toil-level
                    # retry which causes more actual jobs to try to run.
                    # TODO: If there's some kind of permanent failure deleting
                    # from the job store, we will loop on this job forever!
                    self.process_finished_job_description(readyJob, 0)
                else:
                    # No exception during removal. Note that the job is removed.
                    self.processRemovedJob(readyJob, 0)
            else:
                self.processTotallyFailedJob(job_id)
                logger.error("Job: %s is empty but completely failed - something is very wrong", readyJob.jobStoreID)

    def _processReadyJobs(self):
        """Process jobs that are ready to be scheduled/have successors to schedule."""
        logger.debug('Built the jobs list, currently have %i jobs to update and %i jobs issued',
                     self._messages.count(JobUpdatedMessage), self.getNumberOfJobsIssued())

        # Now go through and, for each job that has updated this tick, process it.

        # We are in trouble if the same job is updated with multiple statuses,
        # and we want to coalesce multiple updates with the same status in the
        # same tick, so we remember what we've done so far in this dict from
        # job ID to status.
        handled_with_status = {}
        for message in self._messages.for_each(JobUpdatedMessage):
            # Handle all the job update messages that came in.

            if message.job_id in handled_with_status:
                if handled_with_status[message.job_id] == message.result_status:
                    # This is a harmless duplicate
                    logger.debug("Job %s already updated this tick with status %s and "
                                 "we've received duplicate message %s", message.job_id,
                                 handled_with_status[message.job_id], message)
                else:
                    # This is a conflicting update. We may have already treated
                    # a job as succeeding but now we've heard it's failed, or
                    # visa versa.
                    # This probably shouldn't happen, but does because the
                    # scheduler is not correct somehow and hasn't been for a
                    # long time. Complain about it.
                    logger.warning("Job %s already updated this tick with status %s "
                                   "but we've now received %s", message.job_id,
                                   handled_with_status[message.job_id], message)
                # Either way, we only want to handle one update per tick, like
                # the old dict-based implementation.
                continue
            else:
                # New job for this tick so actually handle that it is updated
                self._processReadyJob(message.job_id, message.result_status)
                handled_with_status[message.job_id] = message.result_status

    def _startServiceJobs(self):
        """Start any service jobs available from the service manager."""
        self.issueQueingServiceJobs()
        while True:
            service_id = self.serviceManager.get_startable_service(0)
            # Stop trying to get jobs when function returns None
            if service_id is None:
                break

            logger.debug('Launching service job: %s', self.toilState.get_job(service_id))
            self.issueServiceJob(service_id)

    def _processJobsWithRunningServices(self):
        """Get jobs whose services have started."""
        while True:
            client_id = self.serviceManager.get_ready_client(0)
            if client_id is None: # Stop trying to get jobs when function returns None
                break
            logger.debug('Job: %s has established its services; all services are running', client_id)

            # Grab the client job description
            client = self.toilState.get_job(client_id)

            # Drop all service relationships
            client.filterServiceHosts(lambda ignored: False)
            self._messages.publish(JobUpdatedMessage(client_id, 0))

    def _processJobsWithFailedServices(self):
        """Get jobs whose services have failed to start."""
        while True:
            client_id = self.serviceManager.get_unservable_client(0)
            if client_id is None: # Stop trying to get jobs when function returns None
                break
            logger.debug('Job: %s has failed to establish its services.', client_id)

            # Grab the client job description
            client = self.toilState.get_job(client_id)

            # Make sure services still want to run
            assert next(client.serviceHostIDsInBatches(), None) is not None

            # Mark the service job updated so we don't stop here.
            self._messages.publish(JobUpdatedMessage(client_id, 1))

    def _gatherUpdatedJobs(self, updatedJobTuple):
        """Gather any new, updated JobDescriptions from the batch system."""
        bsID, exitStatus, exitReason, wallTime = (
            updatedJobTuple.jobID, updatedJobTuple.exitStatus, updatedJobTuple.exitReason,
            updatedJobTuple.wallTime)
        # easy, track different state
        try:
            updatedJob = self.toilState.get_job(self.issued_jobs_by_batch_system_id[bsID])
        except KeyError:
            logger.warning("A result seems to already have been processed for job %s", bsID)
        else:
            if exitStatus == 0:
                logger.debug('Job ended: %s', updatedJob)
            else:
                logger.warning(f'Job failed with exit value {exitStatus}: {updatedJob}\n'
                               f'Exit reason: {exitReason}')
                if exitStatus == CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE:
                    # This is a CWL job informing us that the workflow is
                    # asking things of us that Toil can't do. When we raise an
                    # exception because of this, make sure to forward along
                    # this exit code.
                    logger.warning("This indicates an unsupported CWL requirement!")
                    self.recommended_fail_exit_code = CWL_UNSUPPORTED_REQUIREMENT_EXIT_CODE
            # Tell everyone it stopped running.
            self._messages.publish(JobCompletedMessage(updatedJob.get_job_kind(), updatedJob.jobStoreID, exitStatus))
            self.process_finished_job(bsID, exitStatus, wall_time=wallTime, exit_reason=exitReason)

    def _processLostJobs(self):
        """Process jobs that have gone awry."""
        # In the case that there is nothing happening (no updated jobs to
        # gather for rescueJobsFrequency seconds) check if there are any jobs
        # that have run too long (see self.reissueOverLongJobs) or which have
        # gone missing from the batch system (see self.reissueMissingJobs)
        if ((time.time() - self.timeSinceJobsLastRescued) >= self.config.rescueJobsFrequency):
            # We only rescue jobs every N seconds, and when we have apparently
            # exhausted the current job supply
            self.reissueOverLongJobs()

            hasNoMissingJobs = self.reissueMissingJobs()
            if hasNoMissingJobs:
                self.timeSinceJobsLastRescued = time.time()
            else:
                # This means we'll try again in a minute, providing things are quiet
                self.timeSinceJobsLastRescued += 60

    def innerLoop(self):
        """
        Process jobs.

        This is the leader's main loop.
        """
        self.timeSinceJobsLastRescued = time.time()

        while self._messages.count(JobUpdatedMessage) > 0 or \
              self.getNumberOfJobsIssued() or \
              self.serviceManager.get_job_count():

            if self._messages.count(JobUpdatedMessage) > 0:
                self._processReadyJobs()

            # deal with service-related jobs
            self._startServiceJobs()
            self._processJobsWithRunningServices()
            self._processJobsWithFailedServices()

            # check in with the batch system
            updatedJobTuple = self.batchSystem.getUpdatedBatchJob(maxWait=2)
            if updatedJobTuple is not None:
                # Collect and process all the updates
                self._gatherUpdatedJobs(updatedJobTuple)
                # As long as we are getting updates we definitely can't be
                # deadlocked.
                self.feed_deadlock_watchdog()
            else:
                # If nothing is happening, see if any jobs have wandered off
                self._processLostJobs()

                if self.deadlockThrottler.throttle(wait=False):
                    # Nothing happened this round and it's been long
                    # enough since we last checked. Check for deadlocks.
                    self.checkForDeadlocks()

            # Check on the associated threads and exit if a failure is detected
            self.statsAndLogging.check()
            self.serviceManager.check()
            # the cluster scaler object will only be instantiated if autoscaling is enabled
            if self.clusterScaler is not None:
                self.clusterScaler.check()

            if self.statusThrottler.throttle(wait=False):
                # Time to tell the user how things are going
                self._reportWorkflowStatus()

            if self.kill_throttler.throttle(wait=False):
                if self.jobStore.read_kill_flag():
                    logger.warning("Received kill via job store. Shutting down.")
                    raise KeyboardInterrupt("killed via job store")

            # Make sure to keep elapsed time and ETA up to date even when no jobs come in
            self.progress_overall.update(incr=0)

        logger.debug("Finished the main loop: no jobs left to run.")

        # Consistency check the toil state
        assert self._messages.empty(), f"Pending messages at shutdown: {self._messages}"
        assert self.toilState.successorCounts == {}, f"Jobs waiting on successors at shutdown: {self.toilState.successorCounts}"
        assert self.toilState.successor_to_predecessors == {}, f"Successors pending for their predecessors at shutdown: {self.toilState.successor_to_predecessors}"
        assert self.toilState.service_to_client == {}, f"Services pending for their clients at shutdown: {self.toilState.service_to_client}"
        assert self.toilState.servicesIssued == {}, f"Services running at shutdown: {self.toilState.servicesIssued}"
        # assert self.toilState.jobsToBeScheduledWithMultiplePredecessors # These are not properly emptied yet
        # assert self.toilState.hasFailedSuccessors == set() # These are not properly emptied yet

    def checkForDeadlocks(self):
        """Check if the system is deadlocked running service jobs."""
        totalRunningJobs = len(self.batchSystem.getRunningBatchJobIDs())
        totalServicesIssued = self.serviceJobsIssued + self.preemptibleServiceJobsIssued

        # If there are no updated jobs and at least some jobs running
        if totalServicesIssued >= totalRunningJobs and totalRunningJobs > 0:
            # Collect all running service job store IDs into a set to compare with the deadlock set
            running_service_ids: Set[str] = set()
            for js_id in self.issued_jobs_by_batch_system_id.values():
                job = self.toilState.get_job(js_id)
                if isinstance(job, ServiceJobDescription) and self.serviceManager.is_running(js_id):
                    running_service_ids.add(js_id)

            if len(running_service_ids) > totalRunningJobs:
                # This is too many services.
                # TODO: couldn't more jobs have started since we polled the
                # running job count?
                raise RuntimeError(f"Supposedly running {len(running_service_ids)} services, which is"
                                   f"more than the {totalRunningJobs} currently running jobs overall.")

            # If all the running jobs are active services then we have a potential deadlock
            if len(running_service_ids) == totalRunningJobs:
                # There could be trouble; we are 100% services.
                # See if the batch system has anything to say for itself about its failure to run our jobs.
                message = self.batchSystem.getSchedulingStatusMessage()
                if message is not None:
                    # Prepend something explaining the message
                    message = f"The batch system reports: {message}"
                else:
                    # Use a generic message if none is available
                    message = "Cluster may be too small."


                # See if this is a new potential deadlock
                if self.potentialDeadlockedJobs != running_service_ids:
                    logger.warning(("Potential deadlock detected! All %s running jobs are service jobs, "
                                    "with no normal jobs to use them! %s"), totalRunningJobs, message)
                    self.potentialDeadlockedJobs = running_service_ids
                    self.potentialDeadlockTime = time.time()
                else:
                    # We wait self.config.deadlockWait seconds before declaring the system deadlocked
                    stuckFor = time.time() - self.potentialDeadlockTime
                    if stuckFor >= self.config.deadlockWait:
                        logger.error("We have been deadlocked since %s on these service jobs: %s",
                                     self.potentialDeadlockTime, self.potentialDeadlockedJobs)
                        raise DeadlockException(("The workflow is service deadlocked - all %d running jobs "
                                                 "have been the same active services for at least %s seconds") % (totalRunningJobs, self.config.deadlockWait))
                    else:
                        # Complain that we are still stuck.
                        waitingNormalJobs = self.getNumberOfJobsIssued() - totalServicesIssued
                        logger.warning(("Potentially deadlocked for %.0f seconds. Waiting at most %.0f more seconds "
                                        "for any of %d issued non-service jobs to schedule and start. %s"),
                                       stuckFor, self.config.deadlockWait - stuckFor, waitingNormalJobs, message)
            else:
                # We have observed non-service jobs running, so reset the potential deadlock
                self.feed_deadlock_watchdog()
        else:
            # We have observed non-service jobs running, so reset the potential deadlock.
            self.feed_deadlock_watchdog()

    def feed_deadlock_watchdog(self) -> None:
        """Note that progress has been made and any pending deadlock checks should be reset."""
        if len(self.potentialDeadlockedJobs) > 0:
            # We thought we had a deadlock. Tell the user it is fixed.
            logger.warning("Potential deadlock has been resolved; detected progress")

        self.potentialDeadlockedJobs = set()
        self.potentialDeadlockTime = 0

    def issueJob(self, jobNode: JobDescription) -> None:
        """Add a job to the queue of jobs currently trying to run."""
        # Never issue the same job multiple times simultaneously
        assert jobNode.jobStoreID not in self.toilState.jobs_issued, \
            f"Attempted to issue {jobNode} multiple times simultaneously!"

        workerCommand = [resolveEntryPoint('_toil_worker'),
                         jobNode.jobName,
                         self.jobStoreLocator,
                         jobNode.jobStoreID]

        for context in self.batchSystem.getWorkerContexts():
            # For each context manager hook the batch system wants to run in
            # the worker, serialize and send it.
            workerCommand.append('--context')
            workerCommand.append(base64.b64encode(pickle.dumps(context)).decode('utf-8'))

        # We locally override the command. This shouldn't get persisted back to
        # the job store, or we will detach the job body from the job
        # description. TODO: Don't do it this way! It's weird!
        jobNode.command = ' '.join(workerCommand)

        omp_threads = os.environ.get('OMP_NUM_THREADS') \
            or str(max(1, int(jobNode.cores)))  # make sure OMP_NUM_THREADS is a positive integer

        job_environment = {
            # Set the number of cores used by OpenMP applications
            'OMP_NUM_THREADS': omp_threads,
        }

        # jobBatchSystemID is an int for each job
        jobBatchSystemID = self.batchSystem.issueBatchJob(jobNode, job_environment=job_environment)
        # Record the job by the ID the batch system will use to talk about it with us
        self.issued_jobs_by_batch_system_id[jobBatchSystemID] = jobNode.jobStoreID
        # Record that this job is issued right now and shouldn't e.g. be issued again.
        self.toilState.jobs_issued.add(jobNode.jobStoreID)
        if jobNode.preemptible:
            # len(issued_jobs_by_batch_system_id) should always be greater than or equal to preemptibleJobsIssued,
            # so increment this value after the job is added to the issuedJob dict
            self.preemptibleJobsIssued += 1
        cur_logger = logger.debug if jobNode.local else logger.info
        cur_logger("Issued job %s with job batch system ID: "
                   "%s and %s",
                   jobNode, str(jobBatchSystemID), jobNode.requirements_string())
        # Tell everyone it is issued and the queue size changed
        self._messages.publish(JobIssuedMessage(jobNode.get_job_kind(), jobNode.jobStoreID, jobBatchSystemID))
        self._messages.publish(QueueSizeMessage(self.getNumberOfJobsIssued()))
        # Tell the user there's another job to do
        self.progress_overall.total += 1
        self.progress_overall.update(incr=0)

    def issueJobs(self, jobs):
        """Add a list of jobs, each represented as a jobNode object."""
        for job in jobs:
            self.issueJob(job)

    def issueServiceJob(self, service_id: str) -> None:
        """
        Issue a service job.

        Put it on a queue if the maximum number of service jobs to be scheduled has been reached.
        """
        # Grab the service job description
        service = self.toilState.get_job(service_id)
        assert isinstance(service, ServiceJobDescription)

        if service.preemptible:
            self.preemptibleServiceJobsToBeIssued.append(service_id)
        else:
            self.serviceJobsToBeIssued.append(service_id)
        self.issueQueingServiceJobs()

    def issueQueingServiceJobs(self):
        """Issues any queuing service jobs up to the limit of the maximum allowed."""
        while len(self.serviceJobsToBeIssued) > 0 and self.serviceJobsIssued < self.config.maxServiceJobs:
            self.issueJob(self.toilState.get_job(self.serviceJobsToBeIssued.pop()))
            self.serviceJobsIssued += 1
        while len(self.preemptibleServiceJobsToBeIssued) > 0 and self.preemptibleServiceJobsIssued < self.config.maxPreemptibleServiceJobs:
            self.issueJob(self.toilState.get_job(self.preemptibleServiceJobsToBeIssued.pop()))
            self.preemptibleServiceJobsIssued += 1

    def getNumberOfJobsIssued(self, preemptible: Optional[bool]=None) -> int:
        """
        Get number of jobs that have been added by issueJob(s) and not removed by removeJob.

        :param preemptible: If none, return all types of jobs.
          If true, return just the number of preemptible jobs. If false, return
          just the number of non-preemptible jobs.
        """
        if preemptible is None:
            return len(self.issued_jobs_by_batch_system_id)
        elif preemptible:
            return self.preemptibleJobsIssued
        else:
            assert len(self.issued_jobs_by_batch_system_id) >= self.preemptibleJobsIssued
            return len(self.issued_jobs_by_batch_system_id) - self.preemptibleJobsIssued

    def _getStatusHint(self) -> str:
        """
        Get a short string describing the current state of the workflow for a human.

        Should include number of currently running jobs, number of issued jobs, etc.

        Don't call this too often; it will talk to the batch system, which may
        make queries of the backing scheduler.

        :return: A one-line description of the current status of the workflow.
        """
        # Grab a snapshot of everything. May be inconsistent since it's not atomic.
        issued_job_count = self.getNumberOfJobsIssued()
        running_job_count = len(self.batchSystem.getRunningBatchJobIDs())
        # TODO: When a job stops running but has yet to be collected from the
        # batch system, it will show up here as waiting to run.
        return f"{running_job_count} jobs are running, {issued_job_count - running_job_count} jobs are issued and waiting to run"

    def _reportWorkflowStatus(self) -> None:
        """Report the current status of the workflow to the user."""
        # For now just log our scheduling status message to the log.
        # TODO: make this update fast enough to put it in the progress
        # bar/status line.
        logger.info(self._getStatusHint())

    def removeJob(self, jobBatchSystemID: int) -> JobDescription:
        """
        Remove a job from the system by batch system ID.

        :return: Job description as it was issued.
        """
        assert jobBatchSystemID in self.issued_jobs_by_batch_system_id
        issuedDesc = self.toilState.get_job(self.issued_jobs_by_batch_system_id[jobBatchSystemID])
        if issuedDesc.preemptible:
            # len(issued_jobs_by_batch_system_id) should always be greater than or equal to preemptibleJobsIssued,
            # so decrement this value before removing the job from the issuedJob map
            assert self.preemptibleJobsIssued > 0
            self.preemptibleJobsIssued -= 1
        # It's not issued anymore.
        del self.issued_jobs_by_batch_system_id[jobBatchSystemID]
        assert issuedDesc.jobStoreID in self.toilState.jobs_issued, f"Job {issuedDesc} came back without being issued"
        self.toilState.jobs_issued.remove(issuedDesc.jobStoreID)
        # If service job
        if issuedDesc.jobStoreID in self.toilState.service_to_client:
            # Decrement the number of services
            if issuedDesc.preemptible:
                self.preemptibleServiceJobsIssued -= 1
            else:
                self.serviceJobsIssued -= 1

        # Tell the user that job is done, for progress purposes.
        self.progress_overall.update(incr=1)

        return issuedDesc

    def getJobs(self, preemptible: Optional[bool] = None) -> List[JobDescription]:
        """
        Get all issued jobs.

        :param preemptible: If specified, select only preemptible or only non-preemptible jobs.
        """

        jobs = [self.toilState.get_job(job_store_id) for job_store_id in self.issued_jobs_by_batch_system_id.values()]
        if preemptible is not None:
            jobs = [job for job in jobs if job.preemptible == preemptible]
        return jobs

    def killJobs(self, jobsToKill):
        """
        Kills the given set of jobs and then sends them for processing.

        Returns the jobs that, upon processing, were reissued.
        """
        # If we are rerunning a job we put the ID in this list.
        jobsRerunning = []

        if len(jobsToKill) > 0:
            # Kill the jobs with the batch system. They will now no longer come in as updated.
            self.batchSystem.killBatchJobs(jobsToKill)
            for jobBatchSystemID in jobsToKill:
                # Reissue immediately, noting that we killed the job
                willRerun = self.process_finished_job(jobBatchSystemID, 1, exit_reason=BatchJobExitReason.KILLED)

                if willRerun:
                    # Compose a list of all the jobs that will run again
                    jobsRerunning.append(jobBatchSystemID)

        return jobsRerunning


    #Following functions handle error cases for when jobs have gone awry with the batch system.

    def reissueOverLongJobs(self) -> None:
        """
        Check each issued job.

        If a job is running for longer than desirable issue a kill instruction.
        Wait for the job to die then we pass the job to process_finished_job.
        """
        maxJobDuration = self.config.maxJobDuration
        jobsToKill = []
        if maxJobDuration < 10000000:  # We won't bother doing anything if rescue time > 16 weeks.
            runningJobs = self.batchSystem.getRunningBatchJobIDs()
            for jobBatchSystemID in list(runningJobs.keys()):
                if runningJobs[jobBatchSystemID] > maxJobDuration:
                    logger.warning("The job: %s has been running for: %s seconds, more than the "
                                "max job duration: %s, we'll kill it",
                                self.issued_jobs_by_batch_system_id[jobBatchSystemID],
                                str(runningJobs[jobBatchSystemID]),
                                str(maxJobDuration))
                    jobsToKill.append(jobBatchSystemID)
            reissued = self.killJobs(jobsToKill)
            if len(jobsToKill) > 0:
                # Summarize our actions
                logger.info("Killed %d over long jobs and reissued %d of them", len(jobsToKill), len(reissued))

    def reissueMissingJobs(self, killAfterNTimesMissing=3):
        """
        Check all the current job ids are in the list of currently issued batch system jobs.

        If a job is missing, we mark it as so, if it is missing for a number of runs of
        this function (say 10).. then we try deleting the job (though its probably lost), we wait
        then we pass the job to process_finished_job.
        """
        issuedJobs = set(self.batchSystem.getIssuedBatchJobIDs())
        jobBatchSystemIDsSet = set(list(self.issued_jobs_by_batch_system_id.keys()))
        #Clean up the reissueMissingJobs_missingHash hash, getting rid of jobs that have turned up
        missingJobIDsSet = set(list(self.reissueMissingJobs_missingHash.keys()))
        for jobBatchSystemID in missingJobIDsSet.difference(jobBatchSystemIDsSet):
            self.reissueMissingJobs_missingHash.pop(jobBatchSystemID)
            logger.warning("Batch system id: %s is no longer missing", str(jobBatchSystemID))
        assert issuedJobs.issubset(jobBatchSystemIDsSet) #Assert checks we have
        #no unexpected jobs running
        jobsToKill = []
        for jobBatchSystemID in set(jobBatchSystemIDsSet.difference(issuedJobs)):
            jobStoreID = self.issued_jobs_by_batch_system_id[jobBatchSystemID]
            if jobBatchSystemID in self.reissueMissingJobs_missingHash:
                self.reissueMissingJobs_missingHash[jobBatchSystemID] += 1
            else:
                self.reissueMissingJobs_missingHash[jobBatchSystemID] = 1
            timesMissing = self.reissueMissingJobs_missingHash[jobBatchSystemID]
            logger.warning("Job store ID %s with batch system id %s is missing for the %i time",
                        jobStoreID, str(jobBatchSystemID), timesMissing)
            # Tell everyone it is missing
            self._messages.publish(JobMissingMessage(jobStoreID))
            if timesMissing == killAfterNTimesMissing:
                self.reissueMissingJobs_missingHash.pop(jobBatchSystemID)
                jobsToKill.append(jobBatchSystemID)
        self.killJobs(jobsToKill)
        return len( self.reissueMissingJobs_missingHash ) == 0 #We use this to inform
        #if there are missing jobs

    def processRemovedJob(self, issuedJob, result_status):
        if result_status != 0:
            logger.warning("Despite the batch system claiming failure the "
                        "job %s seems to have finished and been removed", issuedJob)
        self._updatePredecessorStatus(issuedJob.jobStoreID)

    def process_finished_job(self, batch_system_id, result_status, wall_time=None, exit_reason=None) -> bool:
        """
        Process finished jobs.

        Called when an attempt to run a job finishes, either successfully or otherwise.

        Takes the job out of the issued state, and then works out what
        to do about the fact that it succeeded or failed.

        :returns: True if the job is going to run again, and False if the job is
                  fully done or completely failed.
        """
        # De-issue the job.
        issued_job = self.removeJob(batch_system_id)

        if result_status != 0:
            # Show job as failed in progress (and take it from completed)
            self.progress_overall.update(incr=-1)
            self.progress_failed.update(incr=1)

        # Delegate to the vers
        return self.process_finished_job_description(issued_job, result_status, wall_time, exit_reason, batch_system_id)

    def process_finished_job_description(self, finished_job: JobDescription, result_status: int,
                                         wall_time: Optional[float] = None,
                                         exit_reason: Optional[BatchJobExitReason] = None,
                                         batch_system_id: Optional[int] = None) -> bool:
        """
        Process a finished JobDescription based upon its succees or failure.

        If wall-clock time is available, informs the cluster scaler about the
        job finishing.

        If the job failed and a batch system ID is available, checks for and
        reports batch system logs.

        Checks if it succeeded and was removed, or if it failed and needs to be
        set up after failure, and dispatches to the appropriate function.

        :returns: True if the job is going to run again, and False if the job is
                  fully done or completely failed.
        """
        job_store_id = finished_job.jobStoreID
        if wall_time is not None and self.clusterScaler is not None:
            # Tell the cluster scaler a job finished.
            # TODO: Use message bus?
            self.clusterScaler.addCompletedJob(finished_job, wall_time)
        if self.toilState.job_exists(job_store_id):
            logger.debug("Job %s continues to exist (i.e. has more to do)", finished_job)
            try:
                # Reload the job as modified by the worker
                self.toilState.reset_job(job_store_id)
                replacement_job = self.toilState.get_job(job_store_id)
            except NoSuchJobException:
                # We have a ghost job - the job has been deleted but a stale
                # read from e.g. a non-POSIX-compliant filesystem gave us a
                # false positive when we checked for its existence. Process the
                # job from here as any other job removed from the job store.
                # This is a hack until we can figure out how to actually always
                # have a strongly-consistent communications channel. See
                # https://github.com/BD2KGenomics/toil/issues/1091
                logger.warning('Got a stale read for job %s; caught its '
                'completion in time, but other jobs may try to run twice! Fix '
                'the consistency of your job store storage!', finished_job)
                self.processRemovedJob(finished_job, result_status)
                return False
            if replacement_job.logJobStoreFileID is not None:
                with replacement_job.getLogFileHandle(self.jobStore) as log_stream:
                    # more memory efficient than read().striplines() while leaving off the
                    # trailing \n left when using readlines()
                    # http://stackoverflow.com/a/15233739
                    StatsAndLogging.logWithFormatting(job_store_id, log_stream, method=logger.warning,
                                                      message='The job seems to have left a log file, indicating failure: %s' % replacement_job)
                if self.config.writeLogs or self.config.writeLogsGzip:
                    with replacement_job.getLogFileHandle(self.jobStore) as log_stream:
                        StatsAndLogging.writeLogFiles(replacement_job.chainedJobs, log_stream, self.config, failed=True)
            if result_status != 0:
                # If the batch system returned a non-zero exit code then the worker
                # is assumed not to have captured the failure of the job, so we
                # reduce the try count here.
                if replacement_job.logJobStoreFileID is None:
                    logger.warning("No log file is present, despite job failing: %s", replacement_job)

                if batch_system_id is not None:
                    # Look for any standard output/error files created by the batch system.
                    # They will only appear if the batch system actually supports
                    # returning logs to the machine that submitted jobs, or if
                    # --workDir / TOIL_WORKDIR is on a shared file system.
                    # They live directly in the Toil work directory because that is
                    # guaranteed to exist on the leader and workers.
                    file_list = glob.glob(self.batchSystem.format_std_out_err_glob(batch_system_id))
                    for log_file in file_list:
                        try:
                            log_stream = open(log_file, 'rb')
                        except:
                            logger.warning('The batch system left a file %s, but it could not be opened' % log_file)
                        else:
                            with log_stream:
                                if os.path.getsize(log_file) > 0:
                                    StatsAndLogging.logWithFormatting(job_store_id, log_stream, method=logger.warning,
                                                                      message='The batch system left a non-empty file %s:' % log_file)
                                    if self.config.writeLogs or self.config.writeLogsGzip:
                                        file_root, _ = os.path.splitext(os.path.basename(log_file))
                                        job_names = replacement_job.chainedJobs
                                        if job_names is None:   # For jobs that fail this way, replacement_job.chainedJobs is not guaranteed to be set
                                            job_names = [str(replacement_job)]
                                        job_names = [j + '_' + file_root for j in job_names]
                                        log_stream.seek(0)
                                        StatsAndLogging.writeLogFiles(job_names, log_stream, self.config, failed=True)
                                else:
                                    logger.warning('The batch system left an empty file %s' % log_file)

                # Tell the job to reset itself after a failure.
                # It needs to know the failure reason if available; some are handled specially.
                replacement_job.setupJobAfterFailure(exit_status=result_status, exit_reason=exit_reason)
                self.toilState.commit_job(job_store_id)

            elif job_store_id in self.toilState.hasFailedSuccessors:
                # If the job has completed okay, we can remove it from the list of jobs with failed successors
                self.toilState.hasFailedSuccessors.remove(job_store_id)

            # Now that we know the job is done we can add it to the list of updated jobs
            self._messages.publish(JobUpdatedMessage(replacement_job.jobStoreID, result_status))
            logger.debug("Added job: %s to updated jobs", replacement_job)

            # Return True if it will rerun (still has retries) and false if it
            # is completely failed.
            return replacement_job.remainingTryCount > 0
        else:  #The job is done
            self.processRemovedJob(finished_job, result_status)
            # Being done, it won't run again.
            return False

    def getSuccessors(self, job_id: str, alreadySeenSuccessors: Set[str]) -> Set[str]:
        """
        Get successors of the given job by walking the job graph recursively.

        :param alreadySeenSuccessors: any successor seen here is ignored and not traversed.
        :returns: The set of found successors. This set is added to alreadySeenSuccessors.
        """
        successors = set()
        def successorRecursion(job_id: str) -> None:
            # TODO: do we need to reload from the job store here, or is the cache OK?
            jobDesc = self.toilState.get_job(job_id)

            # For lists of successors
            for successorID in jobDesc.allSuccessors():
                # If successor not already visited
                if successorID not in alreadySeenSuccessors:

                    # Add to set of successors
                    successors.add(successorID)
                    alreadySeenSuccessors.add(successorID)

                    # Recurse if job exists
                    # (job may not exist if already completed)
                    if self.toilState.job_exists(successorID):
                        successorRecursion(successorID)

        successorRecursion(job_id)  # Recurse from passed job

        return successors

    def processTotallyFailedJob(self, job_id: str) -> None:
        """Process a totally failed job."""
        # Mark job as a totally failed job
        self.toilState.totalFailedJobs.add(job_id)

        job_desc = self.toilState.get_job(job_id)

        # Tell everyone it failed

        self._messages.publish(JobFailedMessage(job_desc.get_job_kind(), job_id))

        if job_id in self.toilState.service_to_client:
            # Is a service job
            logger.debug("Service job is being processed as a totally failed job: %s", job_desc)

            assert isinstance(job_desc, ServiceJobDescription)

            # Grab the client, which is the predecessor.
            client_id = self.toilState.service_to_client[job_id]

            assert client_id in self.toilState.servicesIssued

            # Leave the service job as a service of its predecessor, because it
            # didn't work.

            # Poke predecessor job that had the service to kill its services,
            # and drop predecessor relationship from ToilState.
            self._updatePredecessorStatus(job_id)

            # Signal to all other services in the group that they should
            # terminate. We do this to prevent other services in the set
            # of services from deadlocking waiting for this service to start
            # properly, and to remember that this service failed with an error
            # and possibly never started.
            if client_id in self.toilState.servicesIssued:
                self.serviceManager.kill_services(self.toilState.servicesIssued[client_id], error=True)
                logger.warning("Job: %s is instructing all other services of its parent job to quit", job_desc)

            # This ensures that the job will not attempt to run any of it's
            # successors on the stack
            self.toilState.hasFailedSuccessors.add(client_id)

            # Remove the start flag, if it still exists. This indicates
            # to the service manager that the job has "started", and prevents
            # the service manager from waiting for it to start forever. It also
            # lets it continue, now that we have issued kill orders for them,
            # to start dependent services, which all need to actually fail
            # before we can finish up with the services' predecessor job.
            self.jobStore.delete_file(job_desc.startJobStoreID)
        else:
            # Is a non-service job
            assert job_id not in self.toilState.servicesIssued
            assert not isinstance(job_desc, ServiceJobDescription)

            # Traverse failed job's successor graph and get the jobStoreID of new successors.
            # Any successor already in toilState.failedSuccessors will not be traversed
            # All successors traversed will be added to toilState.failedSuccessors and returned
            # as a set (unseenSuccessors).
            unseenSuccessors = self.getSuccessors(job_id, self.toilState.failedSuccessors)
            logger.debug("Found new failed successors: %s of job: %s", " ".join(
                         unseenSuccessors), job_desc)

            # For each newly found successor
            for successorJobStoreID in unseenSuccessors:

                # If the successor is a successor of other jobs that have already tried to schedule it
                if successorJobStoreID in self.toilState.successor_to_predecessors:

                    # For each such predecessor job
                    # (we remove the successor from toilState.successor_to_predecessors to avoid doing
                    # this multiple times for each failed predecessor)
                    for predecessor_id in self.toilState.successor_to_predecessors.pop(successorJobStoreID):

                        predecessor = self.toilState.get_job(predecessor_id)

                        # Reduce the predecessor job's successor count.
                        self.toilState.successor_returned(predecessor_id)

                        # Indicate that it has failed jobs.
                        self.toilState.hasFailedSuccessors.add(predecessor_id)
                        logger.debug("Marking job: %s as having failed successors (found by "
                                     "reading successors failed job)", predecessor)

                        # If the predecessor has no remaining successors, add to list of updated jobs
                        if self.toilState.count_pending_successors(predecessor_id) == 0:
                            self._messages.publish(JobUpdatedMessage(predecessor_id, 0))

            # If the job has predecessor(s)
            if job_id in self.toilState.successor_to_predecessors:

                # For each predecessor of the job
                for predecessor_id in self.toilState.successor_to_predecessors[job_id]:

                    # Mark the predecessor as failed
                    self.toilState.hasFailedSuccessors.add(predecessor_id)
                    logger.debug("Totally failed job: %s is marking direct predecessor: %s "
                                 "as having failed jobs", job_desc, self.toilState.get_job(predecessor_id))

                self._updatePredecessorStatus(job_id)

    def _updatePredecessorStatus(self, jobStoreID: str) -> None:
        """Update status of predecessors for finished (possibly failed) successor job."""
        if jobStoreID in self.toilState.service_to_client:
            # Is a service host job, so its predecessor is its client
            client_id = self.toilState.service_to_client.pop(jobStoreID)
            self.toilState.servicesIssued[client_id].remove(jobStoreID)
            if len(self.toilState.servicesIssued[client_id]) == 0: # Predecessor job has
                # all its services terminated
                self.toilState.servicesIssued.pop(client_id) # The job has no running services

                logger.debug('Job %s is no longer waiting on services; all services have stopped', self.toilState.get_job(client_id))

                # Now we know the job is done we can add it to the list of
                # updated job files
                self._messages.publish(JobUpdatedMessage(client_id, 0))
            else:
                logger.debug('Job %s is still waiting on %d services',
                             self.toilState.get_job(client_id),
                             len(self.toilState.servicesIssued[client_id]))
        elif jobStoreID not in self.toilState.successor_to_predecessors:
            #We have reach the root job
            assert self._messages.count(JobUpdatedMessage) == 0, "Root job is done but other jobs are still updated"
            assert len(self.toilState.successor_to_predecessors) == 0, \
                ("Job {} is finished and had no predecessor, but we have other outstanding jobs "
                 "with predecessors: {}".format(jobStoreID, self.toilState.successor_to_predecessors.keys()))
            assert len(self.toilState.successorCounts) == 0, f"Root job is done but jobs waiting on successors: {self.toilState.successorCounts}"
            logger.debug("Reached root job %s so no predecessors to clean up" % jobStoreID)

        else:
            # Is a non-root, non-service job
            logger.debug("Cleaning the predecessors of %s" % jobStoreID)

            # For each predecessor
            for predecessor_id in self.toilState.successor_to_predecessors.pop(jobStoreID):
                assert isinstance(predecessor_id, str), f"Predecessor ID should be str but is {type(predecessor_id)}"
                predecessor = self.toilState.get_job(predecessor_id)

                # Tell the predecessor that this job is done (keep only other successor jobs)
                predecessor.filterSuccessors(lambda jID: jID != jobStoreID)

                # Reduce the predecessor's number of successors by one, as
                # tracked by us, to indicate the completion of the jobStoreID
                # job
                self.toilState.successor_returned(predecessor_id)

                # If the predecessor job is done and all the successors are complete
                if self.toilState.count_pending_successors(predecessor_id) == 0:
                    # Now we know the job is done we can add it to the list of updated job files
                    self._messages.publish(JobUpdatedMessage(predecessor_id, 0))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fcntl
import logging
import os
from abc import abstractmethod
from datetime import datetime
from typing import Dict, Optional, Tuple
from urllib.parse import urlparse

import requests

from toil.lib.io import AtomicFileCreate
from toil.lib.retry import retry

try:
    from toil.lib.aws import get_current_aws_region
    from toil.lib.aws.session import client
    from toil.lib.aws.utils import retry_s3
    HAVE_S3 = True
except ImportError:
    HAVE_S3 = False

logger = logging.getLogger(__name__)

def get_iso_time() -> str:
    """
    Return the current time in ISO 8601 format.
    """
    return datetime.now().isoformat()


def link_file(src: str, dest: str) -> None:
    """
    Create a link to a file from src to dest.
    """
    if os.path.exists(dest):
        raise RuntimeError(f"Destination file '{dest}' already exists.")
    if not os.path.exists(src):
        raise RuntimeError(f"Source file '{src}' does not exist.")
    try:
        os.link(src, dest)
    except OSError:
        os.symlink(src, dest)


def download_file_from_internet(src: str, dest: str, content_type: Optional[str] = None) -> None:
    """
    Download a file from the Internet and write it to dest.
    """
    response = requests.get(src)

    if not response.ok:
        raise RuntimeError("Request failed with a client error or a server error.")

    if content_type and not response.headers.get("Content-Type", "").startswith(content_type):
        val = response.headers.get("Content-Type")
        raise RuntimeError(f"Expected content type to be '{content_type}'.  Not {val}.")

    with open(dest, "wb") as f:
        f.write(response.content)

def download_file_from_s3(src: str, dest: str, content_type: Optional[str] = None) -> None:
    """
    Download a file from Amazon S3 and write it to dest.
    """
    try:
        # Modules to talk to S3 might not always be available so we import things here.
        from toil.lib.aws.utils import get_object_for_url
    except ImportError:
        raise RuntimeError("Cannot access S3 as AWS modules are not available")

    with open(dest, 'wb') as out_stream:
        obj = get_object_for_url(urlparse(src), existing=True)
        obj.download_fileobj(out_stream)

def get_file_class(path: str) -> str:
    """
    Return the type of the file as a human readable string.
    """
    if os.path.islink(path):
        return "Link"
    elif os.path.isfile(path):
        return "File"
    elif os.path.isdir(path):
        return "Directory"
    return "Unknown"

@retry(errors=[OSError, BlockingIOError])
def safe_read_file(file: str) -> Optional[str]:
    """
    Safely read a file by acquiring a shared lock to prevent other processes
    from writing to it while reading.
    """
    try:
        file_obj = open(file)
    except FileNotFoundError:
        return None

    try:
        # acquire a shared lock on the state file, which is blocking until we can lock it
        fcntl.lockf(file_obj.fileno(), fcntl.LOCK_SH)

        try:
            return file_obj.read()
        finally:
            fcntl.flock(file_obj.fileno(), fcntl.LOCK_UN)
    finally:
        file_obj.close()


@retry(errors=[OSError, BlockingIOError])
def safe_write_file(file: str, s: str) -> None:
    """
    Safely write to a file by acquiring an exclusive lock to prevent other
    processes from reading and writing to it while writing.
    """

    if os.path.exists(file):
        # Going to overwrite without anyone else being able to see this
        # intermediate state.

        # Open in read and update mode, so we don't modify the file before we acquire a lock
        file_obj = open(file, "r+")

        try:
            # acquire an exclusive lock
            fcntl.flock(file_obj.fileno(), fcntl.LOCK_EX)

            try:
                file_obj.seek(0)
                file_obj.write(s)
                file_obj.truncate()
            finally:
                fcntl.flock(file_obj.fileno(), fcntl.LOCK_UN)
        finally:
            file_obj.close()
    else:
        # Contend with everyone else to create the file. Last write will win
        # but it will be atomic because of the filesystem.
        with AtomicFileCreate(file) as temp_name:
            with open(temp_name, "w") as file_obj:
                file_obj.write(s)

class MemoryStateCache:
    """
    An in-memory place to store workflow state.
    """

    def __init__(self) -> None:
        """
        Make a new in-memory state cache.
        """

        super().__init__()
        self._data: Dict[Tuple[str, str], Optional[str]] = {}

    def get(self, workflow_id: str, key: str) -> Optional[str]:
        """
        Get a key value from memory.
        """
        return self._data.get((workflow_id, key))

    def set(self, workflow_id: str, key: str, value: Optional[str]) -> None:
        """
        Set or clear a key value in memory.
        """

        if value is None:
            try:
                del self._data[(workflow_id, key)]
            except KeyError:
                pass
        else:
            self._data[(workflow_id, key)] = value

class AbstractStateStore:
    """
    A place for the WES server to keep its state: the set of workflows that
    exist and whether they are done or not.

    This is a key-value store, with keys namespaced by workflow ID. Concurrent
    access from multiple threads or processes is safe and globally consistent.

    Keys and workflow IDs are restricted to [-a-zA-Z0-9_], because backends may
    use them as path or URL components.

    Key values are either a string, or None if the key is not set.

    Workflow existence isn't a thing; nonexistent workflows just have None for
    all keys.

    Note that we don't yet have a cleanup operation: things are stored
    permanently. Even clearing all the keys may leave data behind.

    Also handles storage for a local cache, with a separate key namespace (not
    a read/write-through cache).

    TODO: Can we replace this with just using a JobStore eventually, when
    AWSJobStore no longer needs SimpleDB?
    """

    def __init__(self):
        """
        Set up the AbstractStateStore and its cache.
        """

        # We maintain a local cache here.
        # TODO: Upgrade to an LRU cache wehn we finally learn to paginate
        # workflow status
        self._cache = MemoryStateCache()

    @abstractmethod
    def get(self, workflow_id: str, key: str) -> Optional[str]:
        """
        Get the value of the given key for the given workflow, or None if the
        key is not set for the workflow.
        """
        raise NotImplementedError

    @abstractmethod
    def set(self, workflow_id: str, key: str, value: Optional[str]) -> None:
        """
        Set the value of the given key for the given workflow. If the value is
        None, clear the key.
        """
        raise NotImplementedError

    def read_cache(self, workflow_id: str, key: str) -> Optional[str]:
        """
        Read a value from a local cache, without checking the actual backend.
        """

        return self._cache.get(workflow_id, key)

    def write_cache(self, workflow_id: str, key: str, value: Optional[str]) -> None:
        """
        Write a value to a local cache, without modifying the actual backend.
        """
        self._cache.set(workflow_id, key, value)

class MemoryStateStore(MemoryStateCache, AbstractStateStore):
    """
    An in-memory place to store workflow state, for testing.

    Inherits from MemoryStateCache first to provide implementations for
    AbstractStateStore.
    """

    def __init__(self):
        super().__init__()

class FileStateStore(AbstractStateStore):
    """
    A place to store workflow state that uses a POSIX-compatible file system.
    """

    def __init__(self, url: str) -> None:
        """
        Connect to the state store in the given local directory.

        :param url: Local state store path. Interpreted as a URL, so can't
                    contain ? or #.
        """
        super().__init__()
        parse = urlparse(url)
        if parse.scheme.lower() not in ['file', '']:
            # We want to catch if we get the wrong argument.
            raise RuntimeError(f"{url} doesn't look like a local path")
        if not os.path.exists(parse.path):
            # We need this directory to exist.
            os.makedirs(parse.path, exist_ok=True)
        logger.debug("Connected to FileStateStore at %s", url)
        self._base_dir = parse.path

    def get(self, workflow_id: str, key: str) -> Optional[str]:
        """
        Get a key value from the filesystem.
        """
        return safe_read_file(os.path.join(self._base_dir, workflow_id, key))

    def set(self, workflow_id: str, key: str, value: Optional[str]) -> None:
        """
        Set or clear a key value on the filesystem.
        """
        # Make sure the directory we need exists.
        workflow_dir = os.path.join(self._base_dir, workflow_id)
        os.makedirs(workflow_dir, exist_ok=True)
        file_path = os.path.join(workflow_dir, key)
        if value is None:
            # Delete the file
            try:
                os.unlink(file_path)
            except FileNotFoundError:
                # It wasn't there to start with
                pass
        else:
            # Set the value in the file
            safe_write_file(file_path, value)

if HAVE_S3:
    class S3StateStore(AbstractStateStore):
        """
        A place to store workflow state that uses an S3-compatible object store.
        """

        def __init__(self, url: str) -> None:
            """
            Connect to the state store in the given S3 URL.

            :param url: An S3 URL to a prefix. Interpreted as a URL, so can't
                        contain ? or #.
            """

            super().__init__()

            parse = urlparse(url)

            if parse.scheme.lower() != 's3':
                # We want to catch if we get the wrong argument.
                raise RuntimeError(f"{url} doesn't look like an S3 URL")

            self._bucket = parse.netloc
            # urlparse keeps the leading '/', but here we want a path in the
            # bucket without a leading '/'. We also need to support an empty
            # path.
            self._base_path = parse.path[1:] if parse.path.startswith('/') else parse.path
            self._client = client('s3', region_name=get_current_aws_region())

            logger.debug("Connected to S3StateStore at %s", url)

        def _get_bucket_and_path(self, workflow_id: str, key: str) -> Tuple[str, str]:
            """
            Get the bucket and path in the bucket at which a key value belongs.
            """
            path = os.path.join(self._base_path, workflow_id, key)
            return self._bucket, path

        def get(self, workflow_id: str, key: str) -> Optional[str]:
            """
            Get a key value from S3.
            """
            bucket, path = self._get_bucket_and_path(workflow_id, key)
            for attempt in retry_s3():
                try:
                    logger.debug('Fetch %s path %s', bucket, path)
                    response = self._client.get_object(Bucket=bucket, Key=path)
                    return response['Body'].read().decode('utf-8')
                except self._client.exceptions.NoSuchKey:
                    return None


        def set(self, workflow_id: str, key: str, value: Optional[str]) -> None:
            """
            Set or clear a key value on S3.
            """
            bucket, path = self._get_bucket_and_path(workflow_id, key)
            for attempt in retry_s3():
                if value is None:
                    # Get rid of it.
                    logger.debug('Clear %s path %s', bucket, path)
                    self._client.delete_object(Bucket=bucket, Key=path)
                    return
                else:
                    # Store it, clobbering anything there already.
                    logger.debug('Set %s path %s', bucket, path)
                    self._client.put_object(Bucket=bucket, Key=path,
                                            Body=value.encode('utf-8'))
                    return

# We want to memoize state stores so we can cache on them.
state_store_cache: Dict[str, AbstractStateStore] = {}

def connect_to_state_store(url: str) -> AbstractStateStore:
    """
    Connect to a place to store state for workflows, defined by a URL.

    URL may be a local file path or URL or an S3 URL.
    """

    if url not in state_store_cache:
        # We need to actually make the state store
        parse = urlparse(url)
        if parse.scheme.lower() == 's3':
            # It's an S3 URL
            if HAVE_S3:
                # And we can use S3, so make the right implementation for S3.
                state_store_cache[url] = S3StateStore(url)
            else:
                # We can't actually use S3, so complain.
                raise RuntimeError(f'Cannot connect to {url} because Toil AWS '
                                   f'dependencies are not available. Did you '
                                   f'install Toil with the [aws] extra?')
        elif parse.scheme.lower() in ['file', '']:
            # It's a file URL or path
            state_store_cache[url] = FileStateStore(url)
        else:
            raise RuntimeError(f'Cannot connect to {url} because we do not '
                               f'implement its URL scheme')

    return state_store_cache[url]

class WorkflowStateStore:
    """
    Slice of a state store for the state of a particular workflow.
    """

    def __init__(self, state_store: AbstractStateStore, workflow_id: str) -> None:
        """
        Wrap the given state store for access to the given workflow's state.
        """

        # TODO: We could just use functools.partial on the state store methods
        # to make ours dynamically but that might upset MyPy.
        self._state_store = state_store
        self._workflow_id = workflow_id

    def get(self, key: str) -> Optional[str]:
        """
        Get the given item of workflow state.
        """
        return self._state_store.get(self._workflow_id, key)

    def set(self, key: str, value: Optional[str]) -> None:
        """
        Set the given item of workflow state.
        """
        self._state_store.set(self._workflow_id, key, value)

    def read_cache(self, key: str) -> Optional[str]:
        """
        Read a value from a local cache, without checking the actual backend.
        """

        return self._state_store.read_cache(self._workflow_id, key)

    def write_cache(self, key: str, value: Optional[str]) -> None:
        """
        Write a value to a local cache, without modifying the actual backend.
        """

        self._state_store.write_cache(self._workflow_id, key, value)


def connect_to_workflow_state_store(url: str, workflow_id: str) -> WorkflowStateStore:
    """
    Connect to a place to store state for the given workflow, in the state
    store defined by the given URL.

    :param url: A URL that can be used for connect_to_state_store()
    """

    return WorkflowStateStore(connect_to_state_store(url), workflow_id)

# When we see one of these terminal states, we stay there forever.
TERMINAL_STATES = {"COMPLETE", "EXECUTOR_ERROR", "SYSTEM_ERROR", "CANCELED"}

# How long can a workflow be in CANCELING state before we conclude that the
# workflow running task is gone and move it to CANCELED?
MAX_CANCELING_SECONDS = 30

class WorkflowStateMachine:
    """
    Class for managing the WES workflow state machine.

    This is the authority on the WES "state" of a workflow. You need one to
    read or change the state.

    Guaranteeing that only certain transitions can be observed is possible but
    not worth it. Instead, we just let updates clobber each other and grab and
    cache the first terminal state we see forever. If it becomes important that
    clients never see e.g. CANCELED -> COMPLETE or COMPLETE -> SYSTEM_ERROR, we
    can implement a real distributed state machine here.

    We do handle making sure that tasks don't get stuck in CANCELING.

    State can be:

    "UNKNOWN"
    "QUEUED"
    "INITIALIZING"
    "RUNNING"
    "PAUSED"
    "COMPLETE"
    "EXECUTOR_ERROR"
    "SYSTEM_ERROR"
    "CANCELED"
    "CANCELING"

    Uses the state store's local cache to prevent needing to read things we've
    seen already.
    """

    def __init__(self, store: WorkflowStateStore) -> None:
        """
        Make a new state machine over the given state store slice for the
        workflow.
        """
        self._store = store

    def _set_state(self, state: str) -> None:
        """
        Set the state to the given value, if a read does not show a terminal
        state already.
        We still might miss and clobber transitions to terminal states between
        the read and the write.
        This is not really consistent but also not worth protecting against.
        """

        if self.get_current_state() not in TERMINAL_STATES:
            self._store.set("state", state)

    def send_enqueue(self) -> None:
        """
        Send an enqueue message that would move from UNKNOWN to QUEUED.
        """
        self._set_state("QUEUED")

    def send_initialize(self) -> None:
        """
        Send an initialize message that would move from QUEUED to INITIALIZING.
        """
        self._set_state("INITIALIZING")

    def send_run(self) -> None:
        """
        Send a run message that would move from INITIALIZING to RUNNING.
        """
        self._set_state("RUNNING")

    def send_cancel(self) -> None:
        """
        Send a cancel message that would move to CANCELING from any
        non-terminal state.
        """

        state = self.get_current_state()
        if state != "CANCELING" and state not in TERMINAL_STATES:
            # If it's not obvious we shouldn't cancel, cancel.

            # If we end up in CANCELING but the workflow runner task isn't around,
            # or we signal it at the wrong time, we will stay there forever,
            # because it's responsible for setting the state to anything else.
            # So, we save a timestamp, and if we see a CANCELING status and an old
            # timestamp, we move on.
            self._store.set("cancel_time", get_iso_time())
            # Set state after time, because having the state but no time is an error.
            self._store.set("state", "CANCELING")

    def send_canceled(self) -> None:
        """
        Send a canceled message that would move to CANCELED from CANCELLING.
        """
        self._set_state("CANCELED")

    def send_complete(self) -> None:
        """
        Send a complete message that would move from RUNNING to COMPLETE.
        """
        self._set_state("COMPLETE")

    def send_executor_error(self) -> None:
        """
        Send an executor_error message that would move from QUEUED,
        INITIALIZING, or RUNNING to EXECUTOR_ERROR.
        """
        self._set_state("EXECUTOR_ERROR")

    def send_system_error(self) -> None:
        """
        Send a system_error message that would move from QUEUED, INITIALIZING,
            or RUNNING to SYSTEM_ERROR.
        """
        self._set_state("SYSTEM_ERROR")

    def get_current_state(self) -> str:
        """
        Get the current state of the workflow.
        """

        state = self._store.read_cache("state")
        if state is not None:
            # We permanently cached a terminal state
            return state

        # Otherwise do an actual read from backing storage.
        state = self._store.get("state")

        if state == "CANCELING":
            # Make sure it hasn't been CANCELING for too long.
            # We can get stuck in CANCELING if the workflow-running task goes
            # away or is stopped while reporting back, because it is
            # repsonsible for posting back that it has been successfully
            # canceled.
            canceled_at = self._store.get("cancel_time")
            if canceled_at is None:
                # If there's no timestamp but it's supposedly canceling, put it
                # into SYSTEM_ERROR, because we didn;t move to CANCELING properly.
                state = "SYSTEM_ERROR"
                self._store.set("state", state)
            else:
                # See if it has been stuck canceling for too long
                canceled_at = datetime.fromisoformat(canceled_at)
                canceling_seconds = (datetime.now() - canceled_at).total_seconds()
                if canceling_seconds > MAX_CANCELING_SECONDS:
                    # If it has, go to CANCELED instead, because the task is
                    # nonresponsive and thus not running.
                    state = "CANCELED"
                    self._store.set("state", state)

        if state in TERMINAL_STATES:
            # We can cache this state forever
            self._store.write_cache("state", state)

        if state is None:
            # Make sure we fill in if we couldn't fetch a stored state.
            state = "UNKNOWN"

        return state





import os
import re
import shutil
import subprocess
import tempfile
import sys
import unittest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from toil.test import ToilTest, needs_cwl, needs_docker
from toil.version import python


class ToilDocumentationTest(ToilTest):
    """Tests for scripts in the toil tutorials."""
    @classmethod
    def setUpClass(cls):
        super(ToilTest, cls).setUpClass()
        cls.directory = os.path.dirname(os.path.abspath(__file__))

    def tearDown(self) -> None:
        super(ToilTest, self).tearDown()
        # src/toil/test/docs/scripts/cwlExampleFiles/sample_1_output.txt
        output_files = ["sample_1_output.txt", "sample_2_output.txt", "sample_3_output.txt"]
        for output in output_files:
            output_file = os.path.join(self.directory, 'scripts/cwlExampleFiles', output)
            if os.path.exists(output_file):
                os.remove(output_file)

        jobstores = ['/mnt/ephemeral/workspace/toil-pull-requests/toilWorkflowRun']
        for jobstore in jobstores:
            if os.path.exists(jobstore):
                shutil.rmtree(jobstore)

        unittest.TestCase.tearDown(self)

    """Just check the exit code"""
    def checkExitCode(self, script):
        program = os.path.join(self.directory, "scripts", script)
        process = subprocess.Popen([python, program, "file:my-jobstore", "--clean=always"],
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if isinstance(stdout, bytes):
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')
        if not process.returncode == 0:
            raise RuntimeError(stderr)
        return stdout + ' ' + stderr

    """Check the exit code and the output"""
    def checkExpectedOut(self, script, expectedOutput):
        outerr = self.checkExitCode(script)

        # Check that the expected output is there
        index = outerr.find(expectedOutput)
        self.assertGreater(index, -1, f"Expected:\n{expectedOutput}\nOutput:\n{outerr}")

    """Check the exit code and look for a pattern"""
    def checkExpectedPattern(self, script, expectedPattern):
        outerr = self.checkExitCode(script)

        # Check that the expected output pattern is there
        pattern = re.compile(expectedPattern, re.DOTALL)
        n = re.search(pattern, outerr)
        self.assertNotEqual(n, None, f"Pattern:\n{expectedPattern}\nOutput:\n{outerr}")

    @needs_cwl
    def testCwlexample(self):
        self.checkExitCode("tutorial_cwlexample.py")

    def testDiscoverfiles(self):
        self.checkExitCode("tutorial_discoverfiles.py")

    def testDynamic(self):
        self.checkExitCode("tutorial_dynamic.py")

    def testEncapsulation(self):
        self.checkExitCode("tutorial_encapsulation.py")

    def testEncapsulation2(self):
        self.checkExitCode("tutorial_encapsulation2.py")

    def testHelloworld(self):
        self.checkExpectedOut("tutorial_helloworld.py", "Hello, world!, here's a message: You did it!\n")

    def testInvokeworkflow(self):
        self.checkExpectedOut("tutorial_invokeworkflow.py", "Hello, world!, here's a message: Woot\n")

    def testInvokeworkflow2(self):
        self.checkExpectedOut("tutorial_invokeworkflow2.py", "Hello, world!, I have a message: Woot!\n")

    def testJobFunctions(self):
        self.checkExpectedOut("tutorial_jobfunctions.py", "Hello world, I have a message: Woot!\n")

    def testManaging(self):
        self.checkExitCode("tutorial_managing.py")

    def testManaging2(self):
        self.checkExitCode("tutorial_managing2.py")

    def testMultiplejobs(self):
        self.checkExpectedPattern("tutorial_multiplejobs.py",
                                  "Hello world, I have a message: first.*Hello world, I have a message: "
                                  "second or third.*Hello world, I have a message: second or third.*Hello world,"
                                  " I have a message: last")

    def testMultiplejobs2(self):
        self.checkExpectedPattern("tutorial_multiplejobs2.py",
                                  "Hello world, I have a message: first.*Hello world, I have a message: "
                                  "second or third.*Hello world, I have a message: second or third.*Hello world,"
                                  " I have a message: last")

    def testMultiplejobs3(self):
        self.checkExpectedPattern("tutorial_multiplejobs3.py",
                                  "Hello world, I have a message: first.*Hello world, I have a message: "
                                  "second or third.*Hello world, I have a message: second or third.*Hello world,"
                                  " I have a message: last")

    def testPromises2(self):
        self.checkExpectedOut("tutorial_promises2.py",
                              "['00000', '00001', '00010', '00011', '00100', '00101', '00110', '00111',"
                              " '01000', '01001', '01010', '01011', '01100', '01101', '01110', '01111',"
                              " '10000', '10001', '10010', '10011', '10100', '10101', '10110', '10111',"
                              " '11000', '11001', '11010', '11011', '11100', '11101', '11110', '11111']")

    def testQuickstart(self):
        self.checkExpectedOut("tutorial_quickstart.py", "Hello, world!, here's a message: Woot\n")

    def testRequirements(self):
        self.checkExitCode("tutorial_requirements.py")

    def testArguments(self):
        self.checkExpectedOut("tutorial_arguments.py", "Hello, world!, here's a message: Woot")

    @needs_docker
    def testDocker(self):
        self.checkExitCode("tutorial_docker.py")

    def testPromises(self):
        self.checkExpectedPattern("tutorial_promises.py", "i is: 1.*i is: 2.*i is: 3")

    def testServices(self):
        self.checkExitCode("tutorial_services.py")

    def testStaging(self):
        self.checkExitCode("tutorial_staging.py")

if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, Callable, Dict, List, Tuple, Type

if TYPE_CHECKING:
    from toil.batchSystems.abstractBatchSystem import AbstractBatchSystem

logger = logging.getLogger(__name__)

def aws_batch_batch_system_factory():
    from toil.batchSystems.awsBatch import AWSBatchBatchSystem
    return AWSBatchBatchSystem

def gridengine_batch_system_factory():
    from toil.batchSystems.gridengine import GridEngineBatchSystem
    return GridEngineBatchSystem


def parasol_batch_system_factory():
    from toil.batchSystems.parasol import ParasolBatchSystem
    return ParasolBatchSystem


def lsf_batch_system_factory():
    from toil.batchSystems.lsf import LSFBatchSystem
    return LSFBatchSystem


def single_machine_batch_system_factory():
    from toil.batchSystems.singleMachine import SingleMachineBatchSystem
    return SingleMachineBatchSystem


def mesos_batch_system_factory():
    from toil.batchSystems.mesos.batchSystem import MesosBatchSystem
    return MesosBatchSystem


def slurm_batch_system_factory():
    from toil.batchSystems.slurm import SlurmBatchSystem
    return SlurmBatchSystem

def tes_batch_system_factory():
    from toil.batchSystems.tes import TESBatchSystem
    return TESBatchSystem

def torque_batch_system_factory():
    from toil.batchSystems.torque import TorqueBatchSystem
    return TorqueBatchSystem


def htcondor_batch_system_factory():
    from toil.batchSystems.htcondor import HTCondorBatchSystem
    return HTCondorBatchSystem


def kubernetes_batch_system_factory():
    from toil.batchSystems.kubernetes import KubernetesBatchSystem
    return KubernetesBatchSystem


BATCH_SYSTEM_FACTORY_REGISTRY: Dict[str, Callable[[], Type["AbstractBatchSystem"]]] = {
    'aws_batch'      : aws_batch_batch_system_factory,
    'parasol'        : parasol_batch_system_factory,
    'single_machine' : single_machine_batch_system_factory,
    'grid_engine'    : gridengine_batch_system_factory,
    'lsf'            : lsf_batch_system_factory,
    'mesos'          : mesos_batch_system_factory,
    'slurm'          : slurm_batch_system_factory,
    'tes'            : tes_batch_system_factory,
    'torque'         : torque_batch_system_factory,
    'htcondor'       : htcondor_batch_system_factory,
    'kubernetes'     : kubernetes_batch_system_factory
}
BATCH_SYSTEMS = list(BATCH_SYSTEM_FACTORY_REGISTRY.keys())
DEFAULT_BATCH_SYSTEM = 'single_machine'

def addBatchSystemFactory(key: str, batchSystemFactory: Callable[[], Type['AbstractBatchSystem']]):
    """
    Adds a batch system to the registry for workflow-supplied batch systems.
    """
    BATCH_SYSTEMS.append(key)
    BATCH_SYSTEM_FACTORY_REGISTRY[key] = batchSystemFactory

# We need a snapshot save/restore system for testing. We can't just tamper with
# the globals because module-level globals are their own references, so we
# can't touch this module's global name bindings from a client module.

def save_batch_system_plugin_state() -> Tuple[List[str], Dict[str, Callable[[], Type['AbstractBatchSystem']]]]:
    """
    Return a snapshot of the plugin registry that can be restored to remove
    added plugins. Useful for testing the plugin system in-process with other
    tests.
    """

    snapshot = (list(BATCH_SYSTEMS), dict(BATCH_SYSTEM_FACTORY_REGISTRY))
    return snapshot

def restore_batch_system_plugin_state(snapshot: Tuple[List[str], Dict[str, Callable[[], Type['AbstractBatchSystem']]]]):
    """
    Restore the batch system registry state to a snapshot from
    save_batch_system_plugin_state().
    """

    # We need to apply the snapshot without rebinding the names, because that
    # won't affect modules that imported the names.
    wanted_batch_systems, wanted_registry = snapshot
    BATCH_SYSTEMS.clear()
    BATCH_SYSTEMS.extend(wanted_batch_systems)
    BATCH_SYSTEM_FACTORY_REGISTRY.clear()
    BATCH_SYSTEM_FACTORY_REGISTRY.update(wanted_registry)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import uuid
from typing import Optional

import pytest

from toil.jobStores.aws.jobStore import AWSJobStore
from toil.lib.aws import iam
from toil.lib.aws.utils import create_s3_bucket
from toil.lib.ec2 import establish_boto3_session
from toil.test import ToilTest, needs_aws_s3

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)




class IAMTest(ToilTest):
    """Check that given permissions and associated functions perform correctly"""

    def test_permissions_iam(self):
        granted_perms = {'*': {'Action': ['ec2:*', 'iam:*', 's3:*', 'sdb:*'], 'NotAction': []}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is True
        granted_perms = {'*': {'Action': [], 'NotAction': ['s3:*']}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is True

    def test_negative_permissions_iam(self):
        granted_perms = {'*': {'Action': ['ec2:*', 's3:*', 'sdb:*'], 'NotAction': []}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is False
        granted_perms = {'*': {'Action': [], 'NotAction': ['iam:*', 'ec2:*']}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is False

    def test_wildcard_handling(self):
        assert iam.permission_matches_any("iam:CreateRole", ['iam:Create**']) is True
        assert iam.permission_matches_any("iam:GetUser", ['iam:???????']) is True
        assert iam.permission_matches_any("iam:ListRoleTags", ['iam:*?*Tags']) is True
        assert iam.permission_matches_any("iam:*", ["*"]) is True
        assert iam.permission_matches_any("ec2:*", ['iam:*']) is False




# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import shutil
import signal

from toil.common import Toil
from toil.job import Job
from toil.exceptions import FailedJobsException
from toil.test import ToilTest, slow

logger = logging.getLogger(__name__)


class RestartDAGTest(ToilTest):
    """
    Tests that restarted job DAGs don't run children of jobs that failed in the first run till the
    parent completes successfully in the restart.
    """
    def setUp(self):
        super().setUp()
        self.tempDir = self._createTempDir(purpose='tempDir')
        self.testJobStore = self._getTestJobStorePath()

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.testJobStore)

    @slow
    def testRestartedWorkflowSchedulesCorrectJobsOnFailedParent(self):
        self._testRestartedWorkflowSchedulesCorrectJobs('raise')

    @slow
    def testRestartedWorkflowSchedulesCorrectJobsOnKilledParent(self):
        self._testRestartedWorkflowSchedulesCorrectJobs('kill')

    def _testRestartedWorkflowSchedulesCorrectJobs(self, failType):
        """
        Creates a diamond DAG
            /->passingParent-\
        root                 |-->child
           \\->failingParent--/

        where root and passingParent are guaranteed to pass, while failingParent will fail.
        child should not run on start or restart and we assert that by ensuring that a file create
        iff child is run is not present on the system.

        :param str failType: Does failingParent fail on an assertionError, or is it killed.
        """
        # Specify options
        options = Job.Runner.getDefaultOptions(self.testJobStore)
        options.logLevel = 'DEBUG'
        options.retryCount = 0
        options.clean = "never"

        parentFile = os.path.join(self.tempDir, 'parent')
        childFile = os.path.join(self.tempDir, 'child')

        # Make the first job
        root = Job.wrapJobFn(passingFn)
        passingParent = Job.wrapJobFn(passingFn)
        failingParent = Job.wrapJobFn(failingFn, failType=failType, fileName=parentFile)
        child = Job.wrapJobFn(passingFn, fileName=childFile)

        # define the DAG
        root.addChild(passingParent)
        root.addChild(failingParent)
        passingParent.addChild(child)
        failingParent.addChild(child)

        failReasons = []

        assert not os.path.exists(childFile)

        # Run the test
        for runMode in 'start', 'restart':
            self.errorRaised = None
            try:
                with Toil(options) as toil:
                    if runMode == 'start':
                        toil.start(root)
                    else:
                        toil.restart()
            except Exception as e:
                logger.exception(e)
                self.errorRaised = e
            finally:
                # The processing of an AssertionError and FailedJobsException is the same so we do
                # it together in this finally clause.
                if self.errorRaised is not None:
                    if not os.path.exists(parentFile):
                        failReasons.append('The failing parent file did not exist on toil "%s".'
                                           % runMode)
                    if os.path.exists(childFile):
                        failReasons.append('The child file existed.  i.e. the child was run on '
                                           'toil "%s".' % runMode)
                    if isinstance(self.errorRaised, FailedJobsException):
                        if self.errorRaised.numberOfFailedJobs != 3:
                            failReasons.append('FailedJobsException was raised on toil "%s" but '
                                               'the number of failed jobs (%s) was not 3.'
                                               % (runMode, self.errorRaised.numberOfFailedJobs))
                    elif isinstance(self.errorRaised, AssertionError):
                        failReasons.append('Toil raised an AssertionError instead of a '
                                           'FailedJobsException on toil "%s".' % runMode)
                    else:
                        failReasons.append("Toil raised error: %s" % self.errorRaised)
                    self.errorRaised = None
                    options.restart = True
                else:
                    self.fail('No errors were raised on toil "%s".' % runMode)
        if failReasons:
            self.fail('Test failed for ({}) reasons:\n\t{}'.format(len(failReasons),
                                                               '\n\t'.join(failReasons)))

def passingFn(job, fileName=None):
    """
    This function is guaranteed to pass as it does nothing out of the ordinary.  If fileName is
    provided, it will be created.

    :param str fileName: The name of a file that must be created if provided.
    """
    if fileName is not None:
        # Emulates system touch.
        open(fileName, 'w').close()

def failingFn(job, failType, fileName):
    """
    This function is guaranteed to fail via a raised assertion, or an os.kill

    :param job: Job
    :param str failType: 'raise' or 'kill
    :param str fileName: The name of a file that must be created.
    """
    assert failType in ('raise', 'kill')
    # Use that function to avoid code redundancy
    passingFn(job, fileName)

    if failType == 'raise':
        assert False
    else:
        os.kill(os.getpid(), signal.SIGKILL)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import math
import os
import shlex
import time
from shlex import quote
from typing import Dict, List, Optional

from toil.batchSystems.abstractGridEngineBatchSystem import \
    AbstractGridEngineBatchSystem
from toil.lib.misc import CalledProcessErrorStderr, call_command

logger = logging.getLogger(__name__)


class GridEngineBatchSystem(AbstractGridEngineBatchSystem):

    class Worker(AbstractGridEngineBatchSystem.Worker):
        """
        Grid Engine-specific AbstractGridEngineWorker methods
        """
        def getRunningJobIDs(self):
            times = {}
            with self.runningJobsLock:
                currentjobs = {str(self.batchJobIDs[x][0]): x for x in self.runningJobs}
            stdout = call_command(["qstat"])

            for currline in stdout.split('\n'):
                items = currline.strip().split()
                if items:
                    if items[0] in currentjobs and items[4] == 'r':
                        jobstart = " ".join(items[5:7])
                        jobstart = time.mktime(time.strptime(jobstart, "%m/%d/%Y %H:%M:%S"))
                        times[currentjobs[items[0]]] = time.time() - jobstart

            return times

        def killJob(self, jobID):
            call_command(['qdel', self.getBatchSystemID(jobID)])

        def prepareSubmission(self,
                              cpu: int,
                              memory: int,
                              jobID: int,
                              command: str,
                              jobName: str,
                              job_environment: Optional[Dict[str, str]] = None,
                              gpus: Optional[int] = None):
            # POSIX qsub
            # <https://pubs.opengroup.org/onlinepubs/9699919799.2008edition/utilities/qsub.html>
            # expects a single script argument, which is supposed to be a file.
            # Toil commands usually are not file names but also include
            # arguments. So we split off the arguments like the shell would and
            # hope that the qsub we are using is clever enough to forward along
            # arguments. Otherwise, some qsubs will go looking for the full
            # Toil command string as a file.
            return self.prepareQsub(cpu, memory, jobID, job_environment) + shlex.split(command)

        def submitJob(self, subLine):
            stdout = call_command(subLine)
            output = stdout.split('\n')[0].strip()
            result = int(output)
            return result

        def getJobExitCode(self, sgeJobID):
            """
            Get job exist code, checking both qstat and qacct.  Return None if
            still running.  Higher level should retry on
            CalledProcessErrorStderr, for the case the job has finished and
            qacct result is stale.
            """
            # the task is set as part of the job ID if using getBatchSystemID()
            job, task = (sgeJobID, None)
            if '.' in sgeJobID:
                job, task = sgeJobID.split('.', 1)
            assert task is None, "task ids not currently support by qstat logic below"

            # First try qstat to see if job is still running, if not get the
            # status qacct.  Also, qstat is much faster.
            try:
                call_command(["qstat", "-j", str(job)])
                return None
            except CalledProcessErrorStderr as ex:
                if "Following jobs do not exist" not in ex.stderr:
                    raise

            args = ["qacct", "-j", str(job)]
            if task is not None:
                args.extend(["-t", str(task)])
            stdout = call_command(args)
            for line in stdout.split('\n'):
                if line.startswith("failed") and int(line.split()[1]) == 1:
                    return 1
                elif line.startswith("exit_status"):
                    logger.debug('Exit Status: %r', line.split()[1])
                    return int(line.split()[1])
            return None

        """
        Implementation-specific helper methods
        """
        def prepareQsub(self,
                        cpu: int,
                        mem: int,
                        jobID: int,
                        job_environment: Optional[Dict[str, str]] = None) -> List[str]:
            qsubline = ['qsub', '-V', '-b', 'y', '-terse', '-j', 'y', '-cwd',
                        '-N', 'toil_job_' + str(jobID)]

            environment = self.boss.environment.copy()
            if job_environment:
                environment.update(job_environment)

            if environment:
                qsubline.append('-v')
                qsubline.append(','.join(k + '=' + quote(os.environ[k] if v is None else v)
                                         for k, v in environment.items()))

            reqline = list()
            sgeArgs = os.getenv('TOIL_GRIDENGINE_ARGS')
            if mem is not None:
                memStr = str(mem // 1024) + 'K'
                if not self.boss.config.manualMemArgs:
                    # for UGE instead of SGE; see #2309
                    reqline += ['vf=' + memStr, 'h_vmem=' + memStr]
                elif self.boss.config.manualMemArgs and not sgeArgs:
                    raise ValueError("--manualMemArgs set to True, but TOIL_GRIDGENGINE_ARGS is not set."
                                     "Please set TOIL_GRIDGENGINE_ARGS to specify memory allocation for "
                                     "your system.  Default adds the arguments: vf=<mem> h_vmem=<mem> "
                                     "to qsub.")
            if len(reqline) > 0:
                qsubline.extend(['-hard', '-l', ','.join(reqline)])
            if sgeArgs:
                sgeArgs = sgeArgs.split()
                for arg in sgeArgs:
                    if arg.startswith(("vf=", "h_vmem=", "-pe")):
                        raise ValueError("Unexpected CPU, memory or pe specifications in TOIL_GRIDGENGINE_ARGs: %s" % arg)
                qsubline.extend(sgeArgs)
            # If cpu == 1 (or None) then don't add PE env variable to the qsub command.
            #               This will allow for use of the serial queue for these jobs.
            if (os.getenv('TOIL_GRIDENGINE_PE') is not None) and (cpu is not None) and (cpu > 1) :
                peCpu = int(math.ceil(cpu))
                qsubline.extend(['-pe', os.getenv('TOIL_GRIDENGINE_PE'), str(peCpu)])
            elif (cpu is not None) and (cpu > 1):
                raise RuntimeError("must specify PE in TOIL_GRIDENGINE_PE environment variable when using multiple CPUs. "
                                   "Run qconf -spl and your local documentation for possible values")

            stdoutfile: str = self.boss.format_std_out_err_path(jobID, '$JOB_ID', 'out')
            stderrfile: str = self.boss.format_std_out_err_path(jobID, '$JOB_ID', 'err')
            qsubline.extend(['-o', stdoutfile, '-e', stderrfile])

            return qsubline

    """
    The interface for SGE aka Sun GridEngine.
    """

    @classmethod
    def getWaitDuration(cls):
        return 1



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import signal
import time
import uuid
from threading import Thread

from docker.errors import ContainerError
from toil.common import Toil
from toil.job import Job
from toil.exceptions import FailedJobsException
from toil.lib.docker import (FORGO,
                             RM,
                             STOP,
                             apiDockerCall,
                             containerIsRunning,
                             dockerKill)
from toil.test import ToilTest, needs_docker, slow

logger = logging.getLogger(__name__)


@needs_docker
class DockerTest(ToilTest):
    """
    Tests dockerCall and ensures no containers are left around.
    When running tests you may optionally set the TOIL_TEST_TEMP environment
    variable to the path of a directory where you want temporary test files be
    placed. The directory will be created if it doesn't exist. The path may be
    relative in which case it will be assumed to be relative to the project
    root. If TOIL_TEST_TEMP is not defined, temporary files and directories will
    be created in the system's default location for such files and any temporary
    files or directories left over from tests will be removed automatically
    removed during tear down.
    Otherwise, left-over files will not be removed.
    """
    def setUp(self):
        self.tempDir = self._createTempDir(purpose='tempDir')
        self.dockerTestLogLevel = 'INFO'

    def testDockerClean(self,
                        caching=False,
                        detached=True,
                        rm=True,
                        deferParam=None):
        """
        Run the test container that creates a file in the work dir, and sleeps
        for 5 minutes.
        Ensure that the calling job gets SIGKILLed after a minute, leaving
        behind the spooky/ghost/zombie container. Ensure that the container is
        killed on batch system shutdown (through the deferParam mechanism).
        """
        # We need to test the behaviour of `deferParam` with `rm` and
        # `detached`. We do not look at the case where `rm` and `detached` are
        # both True.  This is the truth table for the different combinations at
        # the end of the test. R = Running, X = Does not exist, E = Exists but
        # not running.
        #              None     FORGO     STOP    RM
        #    rm        X         R         X      X
        # detached     X         R         E      X
        #  Neither     X         R         E      X

        data_dir = os.path.join(self.tempDir, 'data')
        working_dir = os.path.join(self.tempDir, 'working')
        test_file = os.path.join(working_dir, 'test.txt')

        os.makedirs(data_dir, exist_ok=True)
        os.makedirs(working_dir, exist_ok=True)

        options = Job.Runner.getDefaultOptions(os.path.join(self.tempDir,
                                                            'jobstore'))
        options.logLevel = self.dockerTestLogLevel
        options.workDir = working_dir
        options.clean = 'always'
        options.retryCount = 0  # we're expecting the job to fail so don't retry!
        options.caching = caching

        # No base64 logic since it might create a name starting with a `-`.
        container_name = uuid.uuid4().hex
        A = Job.wrapJobFn(_testDockerCleanFn,
                          working_dir,
                          detached,
                          rm,
                          deferParam,
                          container_name)
        try:
            with Toil(options) as toil:
                toil.start(A)
        except FailedJobsException:
            # The file created by spooky_container would remain in the directory
            # and since it was created inside the container, it would have had
            # uid and gid == 0 (root) which may cause problems when docker
            # attempts to clean up the jobstore.
            file_stats = os.stat(test_file)
            assert file_stats.st_gid != 0
            assert file_stats.st_uid != 0

            if (rm and (deferParam != FORGO)) or deferParam == RM or deferParam is None:
                # These containers should not exist
                assert containerIsRunning(container_name) is None, \
                    'Container was not removed.'

            elif deferParam == STOP:
                # These containers should exist but be non-running
                assert containerIsRunning(container_name) == False, \
                    'Container was not stopped.'

            else:
                # These containers will be running
                assert containerIsRunning(container_name) == True, \
                    'Container was not running.'
        finally:
            # Clean up
            try:
                dockerKill(container_name, remove=True)
                os.remove(test_file)
            except:
                pass

    def testDockerClean_CRx_FORGO(self):
        self.testDockerClean(caching=False, detached=False, rm=True,
                             deferParam=FORGO)

    def testDockerClean_CRx_STOP(self):
        self.testDockerClean(caching=False, detached=False, rm=True,
                             deferParam=STOP)

    def testDockerClean_CRx_RM(self):
        self.testDockerClean(caching=False, detached=False, rm=True,
                             deferParam=RM)

    @slow
    def testDockerClean_CRx_None(self):
        self.testDockerClean(caching=False, detached=False, rm=True,
                             deferParam=None)

    @slow
    def testDockerClean_CxD_FORGO(self):
        self.testDockerClean(caching=False, detached=True, rm=False,
                             deferParam=FORGO)

    @slow
    def testDockerClean_CxD_STOP(self):
        self.testDockerClean(caching=False, detached=True, rm=False,
                             deferParam=STOP)

    @slow
    def testDockerClean_CxD_RM(self):
        self.testDockerClean(caching=False, detached=True, rm=False,
                             deferParam=RM)

    @slow
    def testDockerClean_CxD_None(self):
        self.testDockerClean(caching=False, detached=True, rm=False,
                             deferParam=None)

    @slow
    def testDockerClean_Cxx_FORGO(self):
        self.testDockerClean(caching=False, detached=False, rm=False,
                             deferParam=FORGO)

    @slow
    def testDockerClean_Cxx_STOP(self):
        self.testDockerClean(caching=False, detached=False, rm=False,
                             deferParam=STOP)

    @slow
    def testDockerClean_Cxx_RM(self):
        self.testDockerClean(caching=False, detached=False, rm=False,
                             deferParam=RM)

    @slow
    def testDockerClean_Cxx_None(self):
        self.testDockerClean(caching=False, detached=False, rm=False,
                             deferParam=None)

    @slow
    def testDockerClean_xRx_FORGO(self):
        self.testDockerClean(caching=True, detached=False, rm=True,
                             deferParam=FORGO)

    @slow
    def testDockerClean_xRx_STOP(self):
        self.testDockerClean(caching=True, detached=False, rm=True,
                             deferParam=STOP)

    @slow
    def testDockerClean_xRx_RM(self):
        self.testDockerClean(caching=True, detached=False, rm=True,
                             deferParam=RM)

    @slow
    def testDockerClean_xRx_None(self):
        self.testDockerClean(caching=True, detached=False, rm=True,
                             deferParam=None)

    @slow
    def testDockerClean_xxD_FORGO(self):
        self.testDockerClean(caching=True, detached=True, rm=False,
                             deferParam=FORGO)

    @slow
    def testDockerClean_xxD_STOP(self):
        self.testDockerClean(caching=True, detached=True, rm=False,
                             deferParam=STOP)

    @slow
    def testDockerClean_xxD_RM(self):
        self.testDockerClean(caching=True, detached=True, rm=False,
                             deferParam=RM)

    @slow
    def testDockerClean_xxD_None(self):
        self.testDockerClean(caching=True, detached=True, rm=False,
                             deferParam=None)

    @slow
    def testDockerClean_xxx_FORGO(self):
        self.testDockerClean(caching=True, detached=False, rm=False,
                             deferParam=FORGO)

    @slow
    def testDockerClean_xxx_STOP(self):
        self.testDockerClean(caching=True, detached=False, rm=False,
                             deferParam=STOP)

    @slow
    def testDockerClean_xxx_RM(self):
        self.testDockerClean(caching=True, detached=False, rm=False,
                             deferParam=RM)

    @slow
    def testDockerClean_xxx_None(self):
        self.testDockerClean(caching=True, detached=False, rm=False,
                             deferParam=None)

    def testDockerPipeChain(self, caching=False):
        """
        Test for piping API for dockerCall().  Using this API (activated when
        list of argument lists is given as parameters), commands a piped
        together into a chain.
        ex:  parameters=[ ['printf', 'x\n y\n'], ['wc', '-l'] ] should execute:
        printf 'x\n y\n' | wc -l
        """
        options = Job.Runner.getDefaultOptions(os.path.join(self.tempDir, 'jobstore'))
        options.logLevel = self.dockerTestLogLevel
        options.workDir = self.tempDir
        options.clean = 'always'
        options.caching = caching
        A = Job.wrapJobFn(_testDockerPipeChainFn)
        rv = Job.Runner.startToil(A, options)
        logger.info('Container pipeline result: %s', repr(rv))
        rv = rv.decode('utf-8')
        assert rv.strip() == '2'

    def testDockerPipeChainErrorDetection(self, caching=False):
        """
        By default, executing cmd1 | cmd2 | ... | cmdN, will only return an
        error if cmdN fails.  This can lead to all manor of errors being
        silently missed.  This tests to make sure that the piping API for
        dockerCall() throws an exception if non-last commands in the chain fail.
        """
        options = Job.Runner.getDefaultOptions(os.path.join(self.tempDir, 'jobstore'))
        options.logLevel = self.dockerTestLogLevel
        options.workDir = self.tempDir
        options.clean = 'always'
        options.caching = caching
        A = Job.wrapJobFn(_testDockerPipeChainErrorFn)
        rv = Job.Runner.startToil(A, options)
        assert rv is True

    def testNonCachingDockerChain(self):
        self.testDockerPipeChain(caching=True)

    def testNonCachingDockerChainErrorDetection(self):
        self.testDockerPipeChainErrorDetection(caching=True)

    def testDockerLogs(self, stream=False, demux=False):
        """Test for the different log outputs when deatch=False."""

        working_dir = os.path.join(self.tempDir, 'working')
        script_file = os.path.join(working_dir, 'script.sh')
        os.makedirs(working_dir, exist_ok=True)

        options = Job.Runner.getDefaultOptions(os.path.join(self.tempDir, 'jobstore'))
        options.logLevel = self.dockerTestLogLevel
        options.workDir = working_dir
        options.clean = 'always'
        A = Job.wrapJobFn(_testDockerLogsFn,
                          working_dir=working_dir,
                          script_file=script_file,
                          stream=stream,
                          demux=demux)

        try:
            rv = Job.Runner.startToil(A, options)
            assert rv is True
        finally:
            try:
                os.remove(script_file)
            except:
                pass

    def testDockerLogs_Stream(self):
        self.testDockerLogs(stream=True, demux=False)

    def testDockerLogs_Demux(self):
        self.testDockerLogs(stream=False, demux=True)

    def testDockerLogs_Demux_Stream(self):
        self.testDockerLogs(stream=True, demux=True)


def _testDockerCleanFn(job,
                       working_dir,
                       detached=None,
                       rm=None,
                       deferParam=None,
                       containerName=None):
    """
    Test function for test docker_clean.  Runs a container with given flags and
    then dies leaving behind a zombie container.
    :param toil.job.Job job: job
    :param working_dir: See `work_dir=` in :func:`dockerCall`
    :param bool rm: See `rm=` in :func:`dockerCall`
    :param bool detached: See `detached=` in :func:`dockerCall`
    :param int deferParam: See `deferParam=` in :func:`dockerCall`
    :param str containerName: See `container_name=` in :func:`dockerCall`
    """
    def killSelf():
        test_file = os.path.join(working_dir, 'test.txt')
        # Kill the worker once we are sure the docker container is started
        while not os.path.exists(test_file):
            logger.debug('Waiting on the file created by spooky_container.')
            time.sleep(1)
        # By the time we reach here, we are sure the container is running.
        time.sleep(1)
        os.kill(os.getpid(), signal.SIGKILL)

    t = Thread(target=killSelf)
    # Make it a daemon thread so that thread failure doesn't hang tests.
    t.daemon = True
    t.start()
    apiDockerCall(job,
                  image='quay.io/ucsc_cgl/spooky_test',
                  working_dir=working_dir,
                  deferParam=deferParam,
                  containerName=containerName,
                  detach=detached,
                  remove=rm,
                  privileged=True)


def _testDockerPipeChainFn(job):
    """Return the result of a simple pipe chain.  Should be 2."""
    parameters = [['printf', 'x\n y\n'], ['wc', '-l']]
    return apiDockerCall(job,
                         image='quay.io/ucsc_cgl/ubuntu:20.04',
                         parameters=parameters,
                         privileged=True)


def _testDockerPipeChainErrorFn(job):
    """Return True if the command exit 1 | wc -l raises a ContainerError."""
    parameters = [['exit', '1'], ['wc', '-l']]
    try:
        apiDockerCall(job,
                      image='quay.io/ucsc_cgl/spooky_test',
                      parameters=parameters)
    except ContainerError:
        return True
    return False


def _testDockerLogsFn(job,
                      working_dir,
                      script_file,
                      stream=False,
                      demux=False):
    """Return True if the test succeeds. Otherwise Exception is raised."""

    # we write a script file because the redirection operator, '>&2', is wrapped
    # in quotes when passed as parameters.
    import textwrap
    bash_script = textwrap.dedent('''
    #!/bin/bash
    echo hello stdout ;
    echo hello stderr >&2 ;
    echo hello stdout ;
    echo hello stderr >&2 ;
    echo hello stdout ;
    echo hello stdout ;
    ''')

    with open(script_file, 'w') as file:
        file.write(bash_script)

    out = apiDockerCall(job,
                        image='quay.io/ucsc_cgl/ubuntu:20.04',
                        working_dir=working_dir,
                        parameters=[script_file],
                        volumes={working_dir: {'bind': working_dir, 'mode': 'rw'}},
                        entrypoint="/bin/bash",
                        stdout=True,
                        stderr=True,
                        stream=stream,
                        demux=demux)

    # we check the output length because order is not guaranteed.
    if stream:
        if demux:
            # a generator with tuples of (stdout, stderr)
            assert hasattr(out, '__iter__')
            for _ in range(6):
                stdout, stderr = next(out)
                if stdout:
                    # len('hello stdout\n') == 13
                    assert len(stdout) == 13
                elif stderr:
                    assert len(stderr) == 13
                else:
                    assert False
        else:
            # a generator with bytes
            assert hasattr(out, '__iter__')
            for _ in range(6):
                assert len(next(out)) == 13
    else:
        if demux:
            # a tuple of (stdout, stderr)
            stdout, stderr = out
            # len('hello stdout\n' * 4) == 52
            assert len(stdout) == 52
            # len('hello stderr\n' * 2) == 26
            assert len(stderr) == 26
        else:
            # a bytes object
            # len('hello stdout\n' * 4 + 'hello stderr\n' * 2) == 78
            assert len(out) == 78

    return True
